# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
This subpackage provides functionality for collecting and filtering aligned
sequencing reads from a BAM file, determining the alleles they suggest at
a locus, and assesing the evidence for particular variants.

In this subpackage, the records stored in the BAM file are referred to as
"alignments," whereas the term "read" may be more familiar. We use the term
"alignment" for consistency with the SAM specification, and since an
individual read from the sequencer may generate any number of alignments in
the case of chimeric alignments and secondary alignments.
'''

from .util import alignment_key, read_key
from .pileup import Pileup
from .pileup_element import PileupElement
from .pileup_collection import PileupCollection

__all__ = [
    "PileupCollection",
    "Pileup",
    "PileupElement",
    "alignment_key",
    "read_key",
]



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging
import warnings

# Biopython (used by varcode) throws a warning when sequences are compared.
warnings.filterwarnings("ignore", message="Biopython Seq objects")

def configure_logging(args=None):
    if args is not None and args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(
        format="%(asctime)s.%(msecs)d %(levelname)s %(module)s - %(funcName)s:"
        " %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        level=level)




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import functools

from nose.tools import eq_

from varlens.commands import reads

from . import data_path, run_and_parse_csv, cols_concat, temp_file

run = functools.partial(run_and_parse_csv, reads.run)

expected_cols = (
    "source,query_name,reference_start,reference_end,cigarstring").split(',')

def test_basic():
    result = run([
        data_path("CELSR1/bams/bam_0.bam"),
    ])
    eq_(result.shape, (953, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_0.bam"),
        "--is-duplicate",
    ])
    eq_(result.shape, (173, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_0.bam"),
        "--is-read1",
    ])
    eq_(result.shape, (481, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_0.bam"),
        "--is-read2",
    ])
    eq_(result.shape, (472, len(expected_cols)))

def test_loci_filtering():
    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
    ])
    eq_(result.shape, (37053, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22:46930257-46930259"
    ])
    eq_(result.shape, (1795, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22/46930256-46930259"
    ])
    eq_(result.shape, (1795, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22:46930257-46930257"
    ])
    eq_(result.shape, (1753, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22:46930257"
    ])
    eq_(result.shape, (1753, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22/46930256"
    ])
    eq_(result.shape, (1753, len(expected_cols)))

def test_read_filtering():
    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--reference-start", '46932059',
    ])
    eq_(result.shape, (26, len(expected_cols)))

    result = run([
        data_path("CELSR1/bams/bam_5.bam"),
        "--reference-start", '46932059',
        "--query-name-contains", '57841',
    ])
    eq_(result.shape, (1, len(expected_cols)))

def test_round_trip():
    with temp_file(".bam") as out:
        reads.run([
            data_path("CELSR1/bams/bam_5.bam"),
            "--locus", "chr22/46930276",
            "--locus", "chr22/46930256",
            "--out", out,
        ])
        result1 = run([
            out,
        ])
        result2 = run([
            data_path("CELSR1/bams/bam_5.bam"),
            "--locus", "chr22/46930276",
            "--locus", "chr22/46930256",
        ])
        eq_(sorted(cols_concat(result1, expected_cols[1:])),
            sorted(cols_concat(result2, expected_cols[1:])))

def test_round_trip_sam():
    with temp_file(".sam") as out:
        print(out)
        reads.run([
            data_path("CELSR1/bams/bam_5.bam"),
            "--locus", "chr22/46930276",
            "--locus", "chr22/46930256",
            "--out", out,
        ])
        result1 = run([
            out,
        ])
        result2 = run([
            data_path("CELSR1/bams/bam_5.bam"),
            "--locus", "chr22/46930276",
            "--locus", "chr22/46930256",
        ])
        eq_(sorted(cols_concat(result1, expected_cols[1:])),
            sorted(cols_concat(result2, expected_cols[1:])))



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from collections import OrderedDict

class Pileup(object):
    '''
    A Pileup is a collection of PileupElement instances at a particular locus.

    Attributes
    ----------
    locus : Varcode.Locus
        The reference locus. Must be length 1, i.e. a single base.

    elements : OrderedDict of PileupElement instances
        This is logically and ordered set, which we implement as an OrderedDict
        with all values mapping to None.
    '''
    def __init__(self, locus, elements):
        '''
        Construct a new Pileup.

        Parameters
        ----------
        locus : Varcode.Locus
            The reference locus. Must be length 1, i.e. a single base.

        elements : iterable of PileupElement
            The pileup elements. The locus field of these instances must 
            match the locus parameter.
        '''
        self.locus = locus
        self.elements = OrderedDict((e, None) for e in elements)
        assert all(e.locus == self.locus for e in self.elements)

    def __iter__(self):
        return iter(self.elements)

    def __len__(self):
        return len(self.elements)

    def append(self, element):
        '''
        Append a PileupElement to this Pileup. If an identical PileupElement is
        already part of this Pileup, do nothing.
        '''
        assert element.locus == self.locus, (
            "Element locus (%s) != Pileup locus (%s)"
            % (element.locus, self.locus))
        self.elements[element] = None

    def update(self, other):
        '''
        Add all pileup elements from other into self.
        '''
        assert self.locus == other.locus
        self.elements.update(other.elements)

    def filter(self, filters):
        '''
        Apply filters to the pileup elements, and return a new Pileup with the
        filtered elements removed.

        Parameters
        ----------
        filters : list of PileupElement -> bool callables
            A PileupUp element is retained if all filters return True when
            called on it.
        '''
        new_elements = [
            e for e in self.elements
            if all(function(e) for function in filters)]
        return Pileup(self.locus, new_elements)



'''
Given variants from one or more VCF or CSV files, apply filters, add additional
columns, and output to CSV.

Currently we can only output to CSV, not VCF.

A number of useful annotations can be added for each variant by specifying
options of the form '--include-XXX', e.g. '--include-gene'. See detailed help
below.

Examples:

Print basic info for the variants found in two VCF files. Note that variants
found in both files are listed in one row, and the 'sources' column lists
the files each variant was found in:

    %(prog)s test/data/CELSR1/vcfs/vcf_1.vcf test/data/CELSR1/vcfs/vcf_2.vcf

Same as the above but include additional columns giving varcode variant effect
annotations and the genes the variants overlap, and write to a file:

    %(prog)s test/data/CELSR1/vcfs/vcf_1.vcf test/data/CELSR1/vcfs/vcf_2.vcf \\
        --include-effect \\
        --include-gene \\
        --out /tmp/result.csv

Print counts for number of reads supporting reference/variant/other alleles
from the specified BAMs, counting only reads with mapping quality >= 10:

    %(prog)s test/data/CELSR1/vcfs/vcf_1.vcf \\
        --include-read-evidence \\
        --reads test/data/CELSR1/bams/*.bam \\
        --min-mapping-quality 10

'''
from __future__ import absolute_import

import argparse
import sys
import logging

from . import configure_logging
from .. import variant_includes
from .. import variants_util

parser = argparse.ArgumentParser(usage=__doc__)
variants_util.add_args(parser, positional=True)

group = parser.add_argument_group("variant output")

group.add_argument("--no-standard-columns",
    action="store_true", default=False,
    help="Don't write standard columns (genome, contig, start, end, ref, alt)")

group.add_argument("--chunk-rows", metavar="N", type=int,
    help="Write out current results after processing N rows.")

group.add_argument("--limit", metavar="N", type=int,
    help="Process only the first N variants (useful for testing)")

group.add_argument("--columns",
    help="Column separated list of columns to output")

group.add_argument("--rename-column", nargs=2, action="append", default=[],
    metavar="COL",
    help="Rename output column first argument to second. Can be specified "
    "multiple times by repeating the --rename-column option.")

group.add_argument("--out",
    help="Output file. If not specified the CSV is written to stdout.")

group.add_argument('--include-metadata', action="store_true", default=False,
    help="Output variant metadata when loading from VCF (info column, etc).")

for includeable in variant_includes.INCLUDEABLES:
    includeable.add_args(parser)

group.add_argument("-v", "--verbose", action="store_true", default=False)

def run(raw_args=sys.argv[1:]):
    args = parser.parse_args(raw_args)
    configure_logging(args)

    df = variants_util.load_from_args_as_dataframe(args)
    if df is None:
        parser.error("No variants specified.")

    logging.info("Loaded %d variants." % df.shape[0])

    # We run the inverse of the column renames on the input df.
    column_renames = {}
    if args.rename_column:
        column_renames = dict(args.rename_column)
        column_renames_inverse = dict((v, k) for (k, v) in args.rename_column)
        if len(column_renames) != len(column_renames_inverse):
            raise ValueError("Column renames are not 1:1")

        df.columns = [
            column_renames_inverse.get(col, col) for col in df.columns
        ]

    def save(df):
        if column_renames:
            df = df.copy()
            df.columns = [column_renames.get(col, col) for col in df.columns]

        if args.columns:
            columns = [x.strip() for x in args.columns.split(",")]
        else:
            columns = [x for x in df.columns.tolist() if x != "variant"]
            if not args.include_metadata:
                columns = [
                    x for x in columns
                    if not x.startswith("metadata")
                ]
            if args.no_standard_columns:
                columns = [
                    x for x in columns
                    if x not in variants_util.STANDARD_DATAFRAME_COLUMNS
                ]

        df_save = df[columns].copy()
        df_save.interbase_start = df_save.interbase_start.astype(int)
        df_save.interbase_end = df_save.interbase_end.astype(int)

        if args.out is None:
            # Write to stdout.
            df_save.to_csv(sys.stdout, index=False)
        elif args.out.endswith(".csv"):
            df_save.to_csv(args.out, index=False)
            print("Wrote: %s" % args.out)
        else:
            parser.error("Unsupported output file extension: %s" % args.out)

    for includeable in variant_includes.INCLUDEABLES:
        if includeable.requested(args):
            logging.info("Running includeable: %s" % includeable.name)
            instance = includeable.from_args(args)
            for num_rows in instance.compute(df, chunk_rows=args.chunk_rows):
                if args.chunk_rows is not None:
                    save(df)

    if args.chunk_rows is None:
        save(df)
 



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

import pandas
import varcode
import varcode.reference
import logging

from . import util, loci_util
from .locus import Locus
from .read_evidence import pileup_collection

STANDARD_DATAFRAME_COLUMNS = [
    "genome",
    "contig",
    "interbase_start",
    "interbase_end",
    "ref",
    "alt",
]

def add_args(parser, positional=False):
    group = parser.add_argument_group("variant loading")
    group.add_argument("variants" if positional else "--variants",
        nargs=("*" if positional else "+"), default=[],
        help="Path to VCF file. Any number of VCF files may be specified. "
        "CSV files in the format outputted by the varlens-variants tool are "
        "also supported.")
    group.add_argument("--genome",
        help="Genome for the variants (e.g. b37). Required when the genome "
        "cannot be guessed from the metadata in the VCF.")
    group.add_argument("--include-failing-variants",
        action="store_true",
        default=False,
        help="Include variants with a non-PASS filter field.")
    group.add_argument("--variant-source-name", nargs="+",
        help="Names for variant sources. Must specify one name per variant "
        "source. If not specified, the filenames are used.")
    group.add_argument("--max-variants-per-source", type=int,
        metavar="N",
        help="Load at most N variants from each source.")
    group.add_argument("--single-variant", nargs=3, action="append",
        default=[], metavar="X",
        help="Literal variant specified as three arguments: LOCUS REF ALT. "
        "Can be specified any number of times by repeating the "
        "--single-variant option.")

    # Filters
    group = parser.add_argument_group("variant filtering",
        "If multiple filters are specified, the variants must pass *all* "
        "filters. For filtering by loci, any variants that overlap "
        "the specified loci are included.")
    group.add_argument("--ref", nargs="+",
        help="Include only variants where ref is one of the given values.")
    group.add_argument("--alt", nargs="+",
        help="Include only variants where alt is one of the given values.")
    loci_util.add_args(util.PrefixedArgumentParser(group, "variant"))

def load_from_args_as_dataframe(args):
    '''
    Given parsed variant-loading arguments, return a pandas DataFrame.

    If no variant loading arguments are specified, return None.
    '''
    if not args.variants and not args.single_variant:
        return None

    if args.variant_source_name:
        variant_source_names = util.expand(
            args.variant_source_name,
            'variant_source_name',
            'variant source',
            len(args.variants))
    else:
        variant_source_names = util.drop_prefix(args.variants)

    variant_to_sources = collections.defaultdict(list)

    dfs = []
    for i in range(len(args.variants)):
        name = variant_source_names[i]
        prefix = (
            'metadata:' if len(args.variants) == 1 else "metadata:%s:" % name)
        df = load_as_dataframe(
            args.variants[i],
            name=name,
            genome=args.genome,
            max_variants=args.max_variants_per_source,
            only_passing=not args.include_failing_variants,
            metadata_column_prefix=prefix)

        if df.shape[0] == 0:
            logging.warn("No variants loaded from: %s" % args.variants[i])
        else:
            for variant in df.variant:
                variant_to_sources[variant].append(name)
            dfs.append(df)

    if args.single_variant:
        variants = []
        extra_args = {}
        if args.genome:
            extra_args = {
                'ensembl': varcode.reference.infer_genome(args.genome)
            }
        for (locus_str, ref, alt) in args.single_variant:
            locus = Locus.parse(locus_str)
            variant = varcode.Variant(
                    locus.contig,
                    locus.inclusive_start,
                    ref,
                    alt,
                    **extra_args)
            variants.append(variant)
            variant_to_sources[variant].append("commandline")
        dfs.append(variants_to_dataframe(variants))

    df = dfs.pop(0)
    for other_df in dfs:
        df = pandas.merge(
            df,
            other_df,
            how='outer',
            on=["variant"] + STANDARD_DATAFRAME_COLUMNS)

    genomes = df["genome"].unique()
    if len(genomes) > 1:
        raise ValueError(
                "Mixing references is not supported. "
                "Reference genomes: %s" % (", ".join(genomes)))

    df["sources"] = [" ".join(variant_to_sources[v]) for v in df.variant]

    # Apply filters:
    if args.ref:
        df = df.ix[df.ref.isin(args.ref)]
    if args.alt:
        df = df.ix[df.alt.isin(args.alt)]
    loci = loci_util.load_from_args(
        util.remove_prefix_from_parsed_args(args, "variant"))
    if loci is not None:
        df = df.ix[[
            loci.intersects(pileup_collection.to_locus(v))
            for v in df.variant
        ]]
    return df

def load_as_dataframe(
        filename,
        loader=None,
        name=None,
        genome=None,
        max_variants=None,
        only_passing=True,
        metadata_column_prefix=''):

    if name is None:
        name = filename

    if loader is None:
        if (filename.endswith(".vcf") or filename.endswith(".vcf.gz")):
            # Load from VCF
            def loader(filename):
                collection = varcode.load_vcf_fast(
                    filename,
                    genome=genome,
                    max_variants=max_variants,
                    only_passing=only_passing,
                    allow_extended_nucleotides=True)
                return variants_to_dataframe(
                    collection,
                    collection.metadata,
                    metadata_column_prefix=metadata_column_prefix)

        elif (filename.endswith(".csv") or filename.endswith(".csv.gz")):
            # Load from csv
            def loader(filename):
                # Ignores only_passing
                df = pandas.read_csv(filename, nrows=max_variants)
                for column in ['ref', 'alt']:
                    df[column] = df[column].fillna('')
                df["variant"] = [
                    dataframe_row_to_variant(row) for (i, row) in df.iterrows()
                ]
                return df
        else:
            raise ValueError(
                "Unsupported input file extension for variants: %s" % filename)

    df = loader(filename)

    if 'genome' not in df:
        df["genome"] = genome

    df["variant"] = [
        dataframe_row_to_variant(row) for (i, row) in df.iterrows()
    ]
    return df

def variants_to_dataframe(
        variants, metadata=None, metadata_column_prefix=""):
    def record(variant):
        d = {
            'variant': variant,
            'genome': str(variant.reference_name),
            'contig': variant.contig,
            'interbase_start': variant.start - 1,
            'interbase_end': variant.end,
            'ref': variant.ref,
            'alt': variant.alt,
        }
        if metadata:
            for (name, value) in metadata.get(variant, {}).items():
                if name == 'info':
                    for (info_col, value) in value.items():
                        column = '%sinfo:%s' % (
                            metadata_column_prefix, info_col)
                        d[column] = value
                else:
                    d["%s%s" % (metadata_column_prefix, name.lower())] = value
        return d

    df = pandas.DataFrame.from_records([record(v) for v in variants])
    column_indices = dict(
        (column, i) for (i, column) in enumerate(STANDARD_DATAFRAME_COLUMNS))
    columns = sorted(df.columns, key=lambda col: column_indices.get(col, 100))
    return df[columns]

def dataframe_row_to_variant(row):
    return varcode.Variant(
            ensembl=row.genome,
            contig=row.contig,
            start=row.interbase_start + 1,
            ref=row.ref,
            alt=row.alt,
            allow_extended_nucleotides=True)

def dataframe_to_variants(df):
    for column in STANDARD_DATAFRAME_COLUMNS:
        if column not in df:
            raise ValueError("Missing column: %s" % column)

    extra_columns = [
        c for c in df.columns if c not in STANDARD_DATAFRAME_COLUMNS
    ]
    metadata = collections.OrderedDict()
    for (i, row) in df.iterrows():
        variant = dataframe_row_to_variant(row)
        # We ignore the interbase_end field.
        metadata[variant] = dict((c, row[c]) for c in extra_columns)

    return varcode.VariantCollection(metadata.keys(), metadata=metadata)

def load_csv(filename, genome=None):
    # Genome is ignored for now.
    df = pandas.read_csv(filename)
    return dataframe_to_variants(df)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pyensembl
import pysam

from . import read_evidence

class ReadSource(object):
    def __init__(self, name, filename, read_filters=[]):
        self.name = name
        self.filename = filename
        self.handle = pysam.Samfile(filename)
        self.read_filters = read_filters

        self.chromosome_name_map = {}
        for name in self.handle.references:
            normalized = pyensembl.locus.normalize_chromosome(name)
            self.chromosome_name_map[normalized] = name
            self.chromosome_name_map[name] = name

    def index_if_needed(self):
        if self.filename.endswith(".bam") and not self.handle.has_index():
            # pysam strangely requires and index even to iterate through a bam.
            logging.info(
                "Attempting to create BAM index for file: %s" % self.filename)
            samtools_output = pysam.index(self.filename)
            logging.info(
                "Done indexing" + (
                    (": " + samtools_output) if samtools_output else ''))

            # Reopen
            self.handle.close()
            self.handle = pysam.Samfile(self.filename)

    def reads(self, loci=None):
        if loci is None:
            def reads_iterator():
                return self.handle.fetch(until_eof=True)
        elif self.filename.endswith(".sam"):
            # Inefficient.
            chromosome_intervals = {}
            for (contig, intervals) in loci.contigs.items():
                try:
                    chromosome = self.chromosome_name_map[contig]
                except KeyError:
                    logging.warn(
                        "No such contig in bam: %s" % contig)
                    continue
                chromosome_intervals[chromosome] = intervals

            def reads_iterator():
                seen = set()
                for read in self.handle.fetch(until_eof=True):
                    intervals = chromosome_intervals.get(read.reference_name)
                    if not intervals or not intervals.overlaps_range(
                            read.reference_start,
                            read.reference_end):
                        continue
                    key = alignment_key(read)
                    if key not in seen:
                        yield read
                        seen.add(key)
        else:
            self.index_if_needed()

            def reads_iterator():
                seen = set()
                for locus in loci:
                    try:
                        chromosome = self.chromosome_name_map[locus.contig]
                    except KeyError:
                        logging.warn(
                            "No such contig in bam: %s" % locus.contig)
                        continue
                    for read in self.handle.fetch(
                            chromosome,
                            locus.start,
                            locus.end):
                        key = alignment_key(read)
                        if key not in seen:
                            yield read
                            seen.add(key)

        return (
            read for read in reads_iterator()
            if self.read_passes_filters(read))

    def read_passes_filters(self, read):
        return all(read_filter(read) for read_filter in self.read_filters)

    def pileups(self, loci):
        self.index_if_needed()
        collection = read_evidence.PileupCollection.from_bam(self.handle, loci)
        if self.read_filters:
            for (locus, pileup) in collection.pileups.items():
                collection.pileups[locus] = pileup.filter(
                    [lambda element:
                        self.read_passes_filters(element.alignment)])
        return collection

def alignment_key(pysam_alignment_record):
    '''
    Return the identifying attributes of a `pysam.AlignedSegment` instance.
    This is necessary since these objects do not support a useful notion of
    equality (they compare on identify by default).
    '''
    return (
        read_key(pysam_alignment_record),
        pysam_alignment_record.query_alignment_start,
        pysam_alignment_record.query_alignment_end,
    )


def read_key(pysam_alignment_record):
    '''
    Given a `pysam.AlignedSegment` instance, return the attributes identifying
    the *read* it comes from (not the alignment). There may be more than one
    alignment for a read, e.g. chimeric and secondary alignments.
    '''
    return (
        pysam_alignment_record.query_name,
        pysam_alignment_record.is_duplicate,
        pysam_alignment_record.is_read1,
        pysam_alignment_record.is_read2,
    )





# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import argparse

def expand(value, arg_name, input_name, length):
    if value is None or len(value) == 0:
        return [None] * length

    if len(value) == length:
        return value

    if len(value) == 1:
        return value * length

    if length == 1:
        raise ValueError(
            "With only 1 {input_name} specified, each {arg_name} argument "
            "should be length 1. If you are trying to specify multiple filters"
            " to apply consecutively, you should specify the entire argument "
            "multiple times."
            .format(
                arg_name=arg_name,
                input_name=input_name,
                length=length,
                actual=len(value)))

    else:
        raise ValueError(
            "Expected argument {arg_name} to be length 1 (i.e. apply to all "
            "{input_name} inputs) or length {length} (i.e. an individual value"
            " for each of the {length} {input_name} inputs), not {actual}."
            .format(
                arg_name=arg_name,
                input_name=input_name,
                length=length,
                actual=len(value)))


def drop_prefix(strings):
    """
    Removes common prefix from a collection of strings
    """
    strings_without_extensions = [
        s.split(".", 2)[0] for s in strings
    ]

    if len(strings_without_extensions) == 1:
        return [os.path.basename(strings_without_extensions[0])]
    prefix_len = len(os.path.commonprefix(strings_without_extensions))
    result = [string[prefix_len:] for string in strings_without_extensions]
    if len(set(result)) != len(strings):
        # If these operations resulted in a collision, just return the original
        # strings.
        return strings
    return result

class PrefixedArgumentParser(object):
    def __init__(self, wrapped, prefix):
        self.wrapped = wrapped
        self.prefix = prefix

    def add_argument(self, name, *args, **kwargs):
        assert name.startswith("--")
        new_name = "--" + self.prefix + "-" + name[2:]
        self.wrapped.add_argument(new_name, *args, **kwargs)


def remove_prefix_from_parsed_args(args, prefix):
    result = argparse.Namespace()
    for (arg, value) in args._get_kwargs():
        if arg.startswith(prefix + "_"):
            setattr(result, arg[len(prefix + "_"):], value)
    return result






import logging
import time
import collections

import pandas
import numpy
import typechecks

import pyfaidx

from . import sequence_context
from . import mhc_binding
from . import reads_util
from . import support
from . import read_evidence

class Includeable(object):
    columns = None

    @classmethod
    def from_args(cls, args):
        return cls()

    def process_chunk(self, df):
        raise NotImplementedError()

    def compute(self, df, chunk_rows=None):
        assert self.columns
        for column in self.columns:
            if column not in df.columns:
                df[column] = numpy.nan
        rows_to_annotate = pandas.isnull(df[self.columns[0]])
        for column in self.columns[1:]:
            rows_to_annotate = rows_to_annotate | pandas.isnull(df[column])

        while rows_to_annotate.sum() > 0:
            if chunk_rows:
                this_chunk_rows = rows_to_annotate & (
                    rows_to_annotate.cumsum() <= chunk_rows)
            else:
                this_chunk_rows = rows_to_annotate

            num_remaining = rows_to_annotate.sum()
            logging.info("%s: %d / %d (%0.1f%%) remaining. Processing %d rows."
                % (
                    self.name,
                    num_remaining,
                    len(rows_to_annotate),
                    num_remaining * 100.0 / len(rows_to_annotate),
                    this_chunk_rows.sum()))

            rows_to_annotate = rows_to_annotate & (~ this_chunk_rows)
            
            if this_chunk_rows.sum() > 0:
                start = time.time()
                df.ix[this_chunk_rows, self.columns] = self.process_chunk(
                    df.ix[this_chunk_rows].copy())[self.columns]
                logging.info("Processed in %f0.2 sec" % (time.time() - start))
            yield this_chunk_rows.sum()

class Effect(Includeable):
    name = "variant effect annotations"
    columns = ["effect"]

    @staticmethod
    def add_args(parser):
        parser = parser.add_argument_group(Effect.name)
        parser.add_argument("--include-effect",
            action="store_true", default=False,
            help="Include varcode effect annotations")

    @staticmethod
    def requested(args):
        return args.include_effect

    def process_chunk(self, df):
        df["effect"] = [
            v.effects().top_priority_effect().short_description
            for v in df["variant"]
        ]
        return df

class Gene(Includeable):
    name = "gene annotations"
    columns = ["gene"]
    
    @staticmethod
    def add_args(parser):
        parser = parser.add_argument_group(Gene.name)
        parser.add_argument("--include-gene",
            action="store_true", default=False,
            help="Include gene names")

    @staticmethod
    def requested(args):
        return args.include_gene

    def process_chunk(self, df):
        df["gene"] = [
            ' '.join(v.gene_names) if v.gene_names else 'None'
            for v in df.variant
        ]
        return df

class Context(Includeable):
    name = "variant sequence context"
    columns = ["context_5_prime", "context_3_prime", "context_mutation"]
    
    @staticmethod
    def add_args(parser):
        parser = parser.add_argument_group(Context.name)
        parser.add_argument("--include-context",
            action="store_true", default=False,
            help="Include variant sequence context")
        parser.add_argument("--reference",
            help="Path to reference fasta (required for sequence context)")
        parser.add_argument("--context-num-bases", type=int, default=15,
            metavar="N",
            help="Num bases of context to include on each side of the variant")

    @classmethod
    def from_args(cls, args):
        if not args.reference:
            raise ValueError(
                "The --reference argument is required when including context")
        return cls(
            reference=pyfaidx.Fasta(args.reference),
            context_num_bases=args.context_num_bases)

    def __init__(self, reference, context_num_bases):
        self.reference = reference
        self.context_num_bases = context_num_bases

    @staticmethod
    def requested(args):
        return args.include_context

    def process_chunk(self, df):
        context_5_prime = []
        context_3_prime = []
        context_mutation = []
        for variant in df.variant:
            tpl = sequence_context.variant_context(
                self.reference,
                variant.contig,
                variant.start,
                variant.end,
                variant.alt,
                self.context_num_bases)
            context_5_prime.append(tpl[0])
            context_mutation.append(tpl[1])
            context_3_prime.append(tpl[2])

        df["context_5_prime"] = context_5_prime
        df["context_3_prime"] = context_3_prime
        df["context_mutation"] = context_mutation
        return df
    
class MHCBindingAffinity(Includeable):
    name = "MHC binding affinity"
    columns = ["binding_affinity", "binding_allele"]

    noncoding_effects = set([
        "intergenic",
        "intronic",
        "non-coding-transcript",
        "3' UTR",
        "5' UTR",
        "silent",
    ])
    
    @staticmethod
    def add_args(parser):
        parser = parser.add_argument_group(MHCBindingAffinity.name)
        parser.add_argument("--include-mhc-binding",
            action="store_true", default=False,
            help="Include MHC binding (tightest affinity and allele)")
        parser.add_argument("--hla",
            help="Space separated list of MHC alleles, e.g. 'A:02:01 A:02:02'")
        parser.add_argument('--hla-file',
            help="Load HLA types from the specified CSV file. It must have "
            "columns: 'donor' and 'hla'")

    @classmethod
    def from_args(cls, args):
        if bool(args.hla) + bool(args.hla_file) != 1:
            raise ValueError("Must specify exactly one of --hla or --hla-file")
        return cls(
            hla=args.hla,
            hla_dataframe=(
                pandas.read_csv(args.hla_file) if args.hla_file else None))

    @staticmethod
    def string_to_hla_alleles(s):
        return s.replace("'", "").split()

    def __init__(self, hla=None, hla_dataframe=None, donor_to_hla=None):
        """
        Specify exactly one of hla, hla_dataframe, or donor_to_hla.

        Parameters
        -----------
        hla : list of string
            HLA alleles to use for all donors

        hla_dataframe : pandas.DataFrame with columns 'donor' and 'hla'
            DataFrame giving HLA alleles for each donor. The 'hla' column
            should be a space separated list of alleles for that donor.

        donor_to_hla : dict of string -> string list
            Map from donor to HLA alleles for that donor.
        """
        if bool(hla) + (hla_dataframe is not None) + bool(donor_to_hla) != 1:
            raise TypeError(
                "Must specify exactly one of hla, hla_dataframe, donor_to_hla")
        
        self.hla = (
            self.string_to_hla_alleles(hla) if typechecks.is_string(hla)
            else hla)
        self.donor_to_hla = donor_to_hla
        if hla_dataframe is not None:
            self.donor_to_hla = {}
            for (i, row) in hla_dataframe.iterrows():
                if row.donor in self.donor_to_hla:
                    raise ValueError("Multiple rows for donor: %s" % row.donor)
                if pandas.isnull(row.hla):
                    self.donor_to_hla[row.donor] = None
                else:
                    self.donor_to_hla[row.donor] = self.string_to_hla_alleles(
                        row.hla)
        assert self.hla is not None or self.donor_to_hla is not None

    @staticmethod
    def requested(args):
        return args.include_mhc_binding

    def process_chunk(self, df):
        drop_donor = False
        if 'donor' not in df:
            df["donor"] = "DONOR1"
            drop_donor = True
        for donor in df.donor.unique():
            rows = (df.donor == donor)
            if 'effect' in df:
                rows = rows & (~df.effect.isin(self.noncoding_effects))
            sub_df = df.loc[rows]
            alleles = self.hla if self.hla else self.donor_to_hla.get(donor)
            if alleles and sub_df.shape[0] > 0:
                result = mhc_binding.binding_affinities(
                    sub_df.variant, alleles)
                df.loc[rows, "binding_affinity"] = (
                    result["binding_affinity"].values)
                df.loc[rows, "binding_allele"] = (
                    result["binding_allele"].values)
        if drop_donor:
            del df["donor"]
        return df

class ReadEvidence(Includeable):
    name = "read evidence"
    default_column_format = "{source}_count_{allele_group}"
    
    @classmethod
    def add_args(cls, parser):
        group = parser.add_argument_group(cls.name)
        group.add_argument("--include-read-evidence",
            action="store_true", default=False,
            help="Include counts of supporting / contradicting reads")
        group.add_argument("--read-sources-file",
            help="Load paths to BAMs from the given csv file.")
        group.add_argument("--read-sources-id-column",
            default="source_id",
            help="Column to use to join read sources with the variants "
            "dataframe.")
        group.add_argument("--read-sources-column", action="append",
            default=[],
            help="Column containing path to reads (e.g. path to a BAM). Can "
            "be specified any number of times. If not specified, all "
            "columns are used.")
        group.add_argument("--always-prefix-column", action="store_true",
            default=False,
            help="Always prefix the column names with the source name and "
            "count group, even when there is only one of each.")
        group.add_argument("--survive-errors", action="store_true",
            default=False,
            help="If an error is encountered log it and try to continue.")
        
        reads_util.add_args(parser)

    @classmethod
    def from_args(cls, args):
        read_sources = reads_util.load_from_args(args)
        read_sources_df = None
        if args.read_sources_file is not None:
            read_sources_df = pandas.read_csv(
                args.read_sources_file,
                index_col=args.read_sources_id_column)
            if args.read_sources_column:
                read_sources_df = read_sources_df[args.read_sources_column]

        source_names = cls.read_source_names(read_sources, read_sources_df)
        if (args.always_prefix_column or len(source_names) > 1):
            column_format = cls.default_column_format
        else:
            column_format = "{allele_group}"
        return cls(
            read_sources=read_sources,
            read_sources_df=read_sources_df,
            column_format=column_format,
            survive_errors=args.survive_errors)

    def __init__(self,
            read_sources=None,
            read_sources_df=None,
            read_filters=[],
            column_format=default_column_format,
            survive_errors=False):
        """
        
        """
        if sum(x is not None for x in [read_sources, read_sources_df]) != 1:
            raise TypeError(
                "Specify exactly one of read_sources, read_sources_df")
 
        self.read_sources = read_sources
        self.read_sources_df = read_sources_df
        self.read_filters = read_filters
        self.column_format = column_format
        self.survive_errors = survive_errors
        self.set_columns()

    @staticmethod
    def read_source_names(read_sources=None, read_sources_df=None):
        if read_sources is not None:
            return [x.name for x in read_sources]
        return read_sources_df.columns.tolist()
   
    def set_columns(self):
        source_names = self.read_source_names(
            read_sources=self.read_sources,
            read_sources_df=self.read_sources_df)
        assert source_names
        self.columns_dict = collections.OrderedDict()
        for source_name in source_names:
            for allele_group in ["num_alt", "num_ref", "total_depth"]:
                column_name = self.column_name(
                    source_name, allele_group)
                self.columns_dict[column_name] = (
                    source_name, allele_group)
        self.columns = list(self.columns_dict)                
        assert self.columns

    def column_name(self, source, allele_group):
        """
        Parameters
        ----------
        source : string
            name of the ReadSource

        allele_group : string
            one of: num_ref, num_alt, total_depth

        Returns
        ----------
        column name : string
        """
        return self.column_format.format(
            source=source,
            allele_group=allele_group)

    @staticmethod
    def requested(args):
        return args.include_read_evidence

    def process_chunk(self, df):
        if self.read_sources_df is None:
            def rows_and_read_sources():
                all_rows = numpy.ones(df.shape[0], dtype=bool)
                yield (all_rows, self.read_sources)
        else:
            def rows_and_read_sources():
                join_col = self.read_sources_df.index.name
                for join_value in df[join_col].unique():
                    rows = df[join_col] == join_value
                    read_paths = self.read_sources_df.ix[join_value]
                    read_sources = []
                    for (name, filename) in read_paths.iteritems():
                        if pandas.isnull(filename):
                            continue
                        relevant_columns = [
                            col for (col, (source_name, allele_group))
                            in self.columns_dict.items()
                            if source_name == name
                        ]
                        if (~pandas.isnull(df[relevant_columns].values)).all():
                            logging.info(
                                "Skipping source %s (%s) for %s: data exists" %
                                (name, filename, join_value))
                            continue
                        try:
                            read_sources.append(reads_util.load_bam(
                                filename,
                                name=name,
                                read_filters=self.read_filters))
                        except Exception as e:
                            logging.error("Error loading bam: %s in %s" %
                                (str(e), filename))
                            if not self.survive_errors:
                                raise
                            continue

                    if rows.sum() > 0 and read_sources:
                        logging.info(
                            "Processing %s=%s (%d rows, %d read sources)" % (
                                join_col,
                                join_value,
                                rows.sum(),
                                len(read_sources)))
                        yield (rows, read_sources)
                    else:
                        logging.info(
                            "Skipping %s=%s (%d rows, %d read sources)" % (
                                join_col,
                                join_value,
                                rows.sum(),
                                len(read_sources)))

        for (rows, sources) in rows_and_read_sources():
            variants = df.variant[rows]
            counter = collections.Counter(variants)
            duplicate_variants = dict(
                (v, c) for (v, c) in counter.items() if c > 1)
            if duplicate_variants:
                raise ValueError("Duplicate variant(s) for this source: %s" %
                        duplicate_variants)
            variant_loci = sorted(set(
                read_evidence.pileup_collection.to_locus(variant)
                for variant in variants))

            allele_support_df = support.allele_support_df(
                variant_loci, sources)
            assert set(s.name for s in sources) == set(
                allele_support_df.source.unique())
            variant_support_df = support.variant_support(
                variants, allele_support_df)
            assert set(s.name for s in sources) == set(
                variant_support_df.minor_axis)

            for allele_group in ["num_alt", "num_ref", "total_depth"]:
                sub_panel = variant_support_df[allele_group, variants]
                for source_column in sub_panel.columns:
                    dest_column = self.column_name(
                        source_column, allele_group)
                    assert dest_column in self.columns, (
                            "Bad column: %s not in %s" % (
                                dest_column, " ".join(self.columns)))
                    values = sub_panel[source_column].values
                    assert len(values) == rows.sum(), "%d != %d" % (
                        len(values), rows.sum())
                    df.loc[rows, dest_column] = values
        return df

INCLUDEABLES = Includeable.__subclasses__()



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

import pandas
import varcode

CACHED_BINDING_AFFINITIES = {}  # (variant, allele -> nm affinity)
BINDING_PREDICTORS = {}
def binding_affinities(variants, alleles, epitope_lengths=[8, 9, 10, 11]):
    # We import these here so we don't depend on these libraries unless this
    # function is called.
    import mhctools
    import topiary

    for allele in alleles:
        if allele not in BINDING_PREDICTORS:
            BINDING_PREDICTORS[allele] = mhctools.NetMHCpan(
                [allele], epitope_lengths=epitope_lengths)
        predictor = BINDING_PREDICTORS[allele]
        predictions = topiary.predict_epitopes_from_variants(
            varcode.VariantCollection([
                v for v in variants
                if (v, allele) not in CACHED_BINDING_AFFINITIES
            ]),
            predictor,
            ic50_cutoff=float('inf'),
            percentile_cutoff=100)
        if len(predictions) > 0:
            predictions_df = pandas.DataFrame(
                predictions.elements, columns=predictions[0]._fields)
            values = predictions_df.groupby("variant")["value"].min()
            for (variant, value) in zip(values.index, values):
                CACHED_BINDING_AFFINITIES[(variant, allele)] = value
        
    result_df = collections.defaultdict(list)
    for variant in variants:
        (binding_affinity, binding_allele) = min(
            (CACHED_BINDING_AFFINITIES.get((variant, allele), float('nan')),
                allele)
            for allele in alleles)
        if pandas.isnull(binding_affinity):
            binding_allele = None
        result_df["variant"].append(variant)
        result_df["binding_affinity"].append(binding_affinity)
        result_df["binding_allele"].append(binding_allele)

    return pandas.DataFrame(result_df)
    



import argparse

def load_variant_collections_parser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--variants", nargs="+", required=True)
    parser.add_argument("--ensembl-version")
    parser.add_argument("--variant-filter")
    return parser

def load_read_sets_parser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--reads", nargs="+", default=[])
    return parser




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import namedtuple

import pyensembl
import typechecks

class Locus(namedtuple("Locus", "contig start end")):
    '''
    A genomic interval in 0-indexed interbase coordinates.

    See this blog post for a discussion on coordinate systems:
        http://alternateallele.blogspot.com/2012/03/genome-coordinate-conventions.html
    '''

    @property
    def inclusive_start(self):
        return self.start + 1

    @property
    def inclusive_end(self):
        return self.end
    
    @property
    def positions(self):
        '''
        A Python range object giving the bases included in this locus.
        '''
        return range(self.start, self.end)

    @property
    def position(self):
        '''
        If this locus spans a single base, this property gives that position.
        Otherwise, raises a ValueError.
        '''
        if self.end != self.start + 1:
            raise ValueError("Not a single base: %s" % str(self))
        return self.start

    # Factory functions.
    @staticmethod
    def from_inclusive_coordinates(contig, start, end=None):
        '''
        Given coordinates in 1-based coordinates that are inclusive on start
        and end, return a Locus instance. Locus instances are always 0-based
        "interbase" coordinates.
        '''
        typechecks.require_string(contig)
        typechecks.require_integer(start)
        if end is None:
            end = start
        typechecks.require_integer(end)
        contig = pyensembl.locus.normalize_chromosome(contig)
        return Locus(contig, start - 1, end)

    @staticmethod
    def from_interbase_coordinates(contig, start, end=None):
        '''
        Given coordinates in 0-based interbase coordinates, return a Locus
        instance.
        '''
        typechecks.require_string(contig)
        typechecks.require_integer(start)
        if end is None:
            end = start + 1
        typechecks.require_integer(end)
        contig = pyensembl.locus.normalize_chromosome(contig)
        return Locus(contig, start, end)

    @staticmethod
    def parse(string):
        match = re.match(r'(\w+)([:/])(\d+)(-(\d+))?', string)
        if match is None:
            raise ValueError("Couldn't parse locus: %s. "
                "Expected format is: chr5:3332 or chr5:3332-5555 for "
                "inclusive 1-based coordinates and chr5/3331 or "
                "chr5/3331-5554 for half-open 0-based coordinates.")

        (contig, symbol, start, _, maybe_end) = match.groups()
        start = int(start)
        end = int(maybe_end) if maybe_end is not None else None

        if symbol == ":":
            # inclusive coordinatess
            return Locus.from_inclusive_coordinates(contig, start, end)
        else:
            # interbase coordinates
            assert symbol == "/"
            return Locus.from_interbase_coordinates(contig, start, end)


# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from nose.tools import eq_

from varlens.commands import allele_support

from . import data_path, run_and_parse_csv, cols_concat

def run(args):
    logging.info("Running with args: " + ' '.join(args))
    return run_and_parse_csv(allele_support.run, args)

expected_cols = [
    "contig", "interbase_start", "interbase_end", "allele", "count",
]

def test_basic():
    result = run([
        "--reads", data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22:46929963", "chr22:46929964",
    ])
    eq_(cols_concat(result, expected_cols),
        {"22-46929962-46929963-C-60", "22-46929963-46929964-A-81"})

    result = run([
        "--reads", data_path("CELSR1/bams/bam_5.bam"),
        "--locus", "chr22:46929963", "chr22:46929964",
        "--is-reverse"
    ])
    eq_(cols_concat(result, expected_cols),
        {"22-46929962-46929963-C-37", "22-46929963-46929964-A-47"})

    result = run([
        "--reads", data_path("gatk_mini_bundle_extract.bam"),
        "--locus", "chr20:10008951",
        "--is-reverse",
    ])
    eq_(cols_concat(result, expected_cols),
        {"20-10008950-10008951-C-1"})

def test_simple():
    result = run([
        "--reads", data_path("CELSR1/bams/bam_0.bam"),
        "--genome", "b37",
        "--variants", data_path("CELSR1/vcfs/vcf_1.vcf"),
    ])
    eq_(cols_concat(
            result,
            ["contig", "interbase_start", "interbase_end", "allele", "count"]),
        {
            '22-50636217-50636218-N-0',
            '22-50875932-50875933-N-0',
            '22-21829554-21829555-N-0',
            "22-46931059-46931060-A-50",
            "22-46931061-46931062-G-51",
    })

    pick_one_variant = [
        ["--ref", "G"],
        ["--alt", "A"],
        ["--variant-locus", "22/46931061"],
        ["--variant-locus", "22/46931061-46931062"],
    ]
    for variant_filter in pick_one_variant:
        result = run([
            "--reads", data_path("CELSR1/bams/bam_0.bam"),
            "--genome", "b37",
            "--variants", data_path("CELSR1/vcfs/vcf_1.vcf"),
        ] + variant_filter)
        yield (
            eq_,
            cols_concat(
                result, ["contig", "interbase_start", "interbase_end"]),
            {"22-46931061-46931062"})




'''
Filter reads from one or more BAMs and output a CSV or a new BAM.

Loci and VCF files may be specified, in which case reads are filtered to
overlap the specified loci or variants.

Examples:

Print basic fields for the reads in a BAM:

    %(prog)s test/data/CELSR1/bams/bam_0.bam

Same as above but filter only to reads aligned on the (-) strand, write to a 
file instead of stdout, and also include the mapping quality and sequenced
bases in the output:

    %(prog)s test/data/CELSR1/bams/bam_0.bam \\
        --is-reverse \\
        --field mapping_quality query_alignment_sequence \\
        --out /tmp/result.csv

Write a bam file consisting of reads with mapping quality >=30 and
overlapping a certain locus:

    %(prog)s test/data/CELSR1/bams/bam_0.bam \\
        --min-mapping-quality 30 \\
        --locus 22:46932040-46932050 \\
        --out /tmp/result.bam

Write a bam file consisting of reads overlapping variants from a VCF:

    %(prog)s test/data/CELSR1/bams/bam_0.bam \\
        --variants test/data/CELSR1/vcfs/vcf_1.vcf \\
        --out /tmp/result.bam

Print just the header for a BAM in csv format:

    %(prog)s test/data/CELSR1/bams/bam_0.bam --header

'''
from __future__ import absolute_import

import argparse
import sys
import csv

import pysam

from . import configure_logging
from .. import loci_util
from .. import reads_util
from .. import variants_util
from ..read_evidence.pileup_collection import PileupCollection, to_locus

STANDARD_FIELDS = [
    "source",
    "query_name",
    "reference_start",
    "reference_end",
    "cigarstring",
]

parser = argparse.ArgumentParser(usage=__doc__)
group = parser.add_argument_group("output")
group.add_argument("--out",
    help="Output file. Format is guessed from file extension: must be csv or "
    "bam. If not specified, csv is written to stdout.")
group.add_argument("--field", nargs="+", default=[],
    help="Additional read fields to output as columns in the csv. See pysam "
    "documentation (http://pysam.readthedocs.org/en/latest/api.html) for the "
    "meaning of these fields. Valid fields include: %s" % (
        " ".join(PileupCollection._READ_ATTRIBUTE_NAMES)))

group.add_argument("--no-standard-fields", action="store_true", default=False,
    help="Do not include the standard fields (%s) in csv output."
    % ', '.join(STANDARD_FIELDS))
group.add_argument("--no-sort", action="store_true", default=False,
    help="When outputting a bam, do not call samtools sort.")
group.add_argument(
    "--header",
    action="store_true",
    default=False,
    help="Output BAM/SAM header only.")
group.add_argument(
    "--header-set",
    nargs=4,
    action="append",
    help="When outputting a bam, set a particular header field to the given "
    "value. Example: --header-set RG . SM my_sample")

group.add_argument("-v", "--verbose", action="store_true", default=False)

reads_util.add_args(parser, positional=True)
loci_util.add_args(parser.add_argument_group("loci specification"))
variants_util.add_args(parser)

def run(raw_args=sys.argv[1:]):
    args = parser.parse_args(raw_args)
    configure_logging(args)

    read_sources = reads_util.load_from_args(args)
    if not read_sources:
        parser.error("No read sources specified.")

    loci = loci_util.load_from_args(args)  # may be None
    variants_df = variants_util.load_from_args_as_dataframe(args)
    if variants_df is not None:
        variant_loci = loci_util.Loci(
            to_locus(variant)
            for variant in variants_df["variant"])
        loci = variant_loci if loci is None else loci.union(variant_loci)

    if args.header:
        if loci is not None:
            parser.error("If specifying --header don't specify loci.")
        if args.field:
            parser.error("If specifying --header don't specify fields.")

    out_pysam_handle = None
    out_csv_writer = out_csv_fd = None
    if args.out and (args.out.endswith(".bam") or args.out.endswith(".sam")):
        if args.field:
            parser.error("Don't specify fields when outputting to bam or sam.")

        header = update_header(args, read_sources[0].handle.header)
        out_pysam_handle = pysam.AlignmentFile(
            args.out,
            "wb",
            header=header)

    elif not args.out or args.out.endswith(".csv"):
        out_csv_fd = open(args.out, "w") if args.out else sys.stdout
        out_csv_writer = csv.writer(out_csv_fd)

        if args.header:
            if args.field:
                parser.error("Don't specify fields when outputting header.")
            out_csv_writer.writerow([
                "read_source", "group", "index", "key", "value",
            ])
        else:
            columns = (
                ([] if args.no_standard_fields else STANDARD_FIELDS) +
                args.field)
            out_csv_writer.writerow(columns)
    else:
        parser.error(
            "Don't know how to write to file with output extension: %s. "
            "Supported extensions: csv, bam, sam." % args.out)

    num_reads = 0
    for read_source in read_sources:
        if args.header:
            header = update_header(args, read_source.handle.header)
            for (group, i, key, value) in reads_util.flatten_header(header):
                out_csv_writer.writerow(
                    [read_source.name, group, str(i), key, value])
            continue  # we don't look at reads at all.
        for read in read_source.reads(loci):
            num_reads += 1
            if out_pysam_handle is not None:
                out_pysam_handle.write(read)
            if out_csv_writer is not None:
                out_csv_writer.writerow([
                    str(read_field(read_source, read, field))
                    for field in columns
                ])

    if out_pysam_handle is not None:
        out_pysam_handle.close()
        if not args.no_sort:
            print("Sorting read file %s" % args.out)
            pysam.sort(
                "-o", args.out,
                "-T", "varlens_reads", args.out,
                catch_stdout=False)
        print("Wrote %d reads: %s" % (num_reads, args.out))

    if out_csv_fd is not None and out_csv_fd is not sys.stdout:
        out_csv_fd.close()
        print("Wrote: %s" % args.out)


def read_field(read_source, read, field_name):
    if field_name == 'source':
        return read_source.name

    if field_name.startswith("tag:"):
        tag_name = field_name[len("tag:"):]
        return read.get_tags().get(tag_name)

    try:
        return getattr(read, field_name)
    except AttributeError:
        raise ValueError("Invalid read field '%s'. Valid fields include: %s"
            % (field_name, ' '.join(dir(read))))

def update_header(args, header):
    if args.header_set:
        header = dict(header)
        for (group, index_string, key, value) in args.header_set:
            if not isinstance(header[group], list):
                header[group] = [header[group]]
            if index_string == ".":
                indices = range(len(header[group]))
            else:
                indices = [int(x) for x in index_string.split(",")]
            for index in indices:
                header[group][index][key] = value
    return header



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import intervaltree

from .locus import Locus

def add_args(parser):
    # TODO:
    # - Load intervals_list files
    parser.add_argument('--locus', nargs="+", default=[],
        help="Genomic locus, like chr1:2342332 or chr1:2342-23423. "
        "Any number of loci may be specified.")
    # parser.add_argument("--neighbor-offsets",
    #    nargs="+", type=int, default=[],
    #    help="")

def load_from_args(args):
    """
    Return a Loci object giving the loci specified on the command line.

    If no loci-related arguments are specified, return None. This makes it
    possible to distinguish an empty set of loci, for example due to filters
    removing all loci, from the case where the user didn't specify any
    arguments.
    """
    if not args.locus:
        return None

    loci_iterator = (Locus.parse(locus) for locus in args.locus)

#   if args.neighbor_offsets:
#       loci_iterator = expand_with_neighbors(
#           loci_iterator, args.neighbor_offsets)

    return Loci(loci_iterator)

# def expand_with_neighbors(loci_iterator, neighbor_offsets):
#    offsets = sorted(set(neighbor_offsets + [0]))
#    for locus in loci_iterator:
#        for offset in offsets:
#            if offset == 0:
#                yield locus
#            else:
#                yield Locus(
#                    locus.contig, locus.start + offset, locus.end + offset)

class Loci(object):
    def __init__(self, locus_iterator=[], contig_map=None):
        self.contigs = collections.defaultdict(intervaltree.IntervalTree)
        if contig_map:
            self.contigs.update(contig_map)
        for locus in locus_iterator:
            self.contigs[locus.contig].addi(locus.start, locus.end)

    def __iter__(self):
        for contig in sorted(self.contigs):
            for interval in self.contigs[contig]:
                yield Locus(contig, interval.begin, interval.end)

    def __len__(self):
        return sum(len(tree) for tree in self.contigs.values())

    def intersects(self, locus):
        return self.contigs[locus.contig].overlaps(locus.start, locus.end)

    def union(self, other):
        contig_map = {}
        for contig in set(self.contigs).union(other.contigs):
            contig_map[contig] = self.contigs[contig].union(
                other.contigs[contig])
        return Loci(contig_map=contig_map)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import subprocess
import warnings
import logging

import pandas
from nose.tools import eq_

from varlens.commands import variants

from . import data_path, run_and_parse_csv, cols_concat, temp_file

def run(args):
    logging.info("Running with args: " + ' '.join(args))
    return run_and_parse_csv(variants.run, args)

reference_fasta = data_path("chr22.no_line_wrap.fa")

expected_cols = [
    "genome", "contig", "interbase_start", "interbase_end", "ref", "alt",
]

def test_basic():
    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--genome", "b37",
    ])
    eq_(sorted(cols_concat(result, expected_cols)), sorted({
        "GRCh37-22-46931059-46931060-A-C",
        "GRCh37-22-21829554-21829555-T-G",
        "GRCh37-22-46931061-46931062-G-A",
        "GRCh37-22-50636217-50636218-A-C",
        "GRCh37-22-50875932-50875933-A-C",
    }))

def test_genes_and_effects():
    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--genome", "b37",
        "--include-effect",
        "--include-gene",
        "--rename-column", "gene", "genez",
    ])
    eq_(sorted(cols_concat(result, expected_cols + ["effect", "genez"])),
        sorted({
           'GRCh37-22-21829554-21829555-T-G-non-coding-transcript-PI4KAP2',
           'GRCh37-22-46931059-46931060-A-C-p.S670A-CELSR1',
           'GRCh37-22-46931061-46931062-G-A-p.S669F-CELSR1',
           'GRCh37-22-50636217-50636218-A-C-intronic-TRABD',
           'GRCh37-22-50875932-50875933-A-C-splice-acceptor-PPP6R2',
        }))

def test_context():
    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--genome", "b37",
        "--include-context",
        "--context-num-bases", "5",
        "--reference", reference_fasta,
    ])
    eq_(sorted(cols_concat(result,
            expected_cols + [
                "context_5_prime", "context_3_prime", "context_mutation"])),
        sorted({
            "GRCh37-22-46931059-46931060-A-C-GCTCC-CCACC-T>G",
            "GRCh37-22-21829554-21829555-T-G-CATGA-AGTGA-T>G",
            "GRCh37-22-46931061-46931062-G-A-GAGCT-CTCCA-C>T",
            "GRCh37-22-50636217-50636218-A-C-AGGGA-GGGCA-T>G",
            "GRCh37-22-50875932-50875933-A-C-AGGCC-GGGAG-T>G",
        }))

def test_mhc_binding_affinity():
    # If netMHC is not installed, we skip this test
    try:
        # If this succeeds (no exception), we do nothing.
        subprocess.call(
            "netMHC", stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        warnings.warn("netMHC not installed, skipping mhc binding test")
        return

    with temp_file(".csv") as out_csv:
        run([
            data_path("CELSR1/vcfs/vcf_1.vcf"),
            "--genome", "b37",
            "--include-mhc-binding",
            "--hla", "A:02:01 A:02:02",
            "--out", out_csv,
        ])
        eq_(sorted(cols_concat(pandas.read_csv(out_csv),
                expected_cols + ["binding_affinity", "binding_allele"])),
            sorted({
               'GRCh37-22-21829554-21829555-T-G-nan-nan',
               'GRCh37-22-46931059-46931060-A-C-142.13-A:02:02',
               'GRCh37-22-46931061-46931062-G-A-115.77-A:02:02',
               'GRCh37-22-50636217-50636218-A-C-nan-nan',
               'GRCh37-22-50875932-50875933-A-C-nan-nan',
            }))

def test_read_evidence():
    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--include-read-evidence",
        "--reads", data_path("CELSR1/bams/bam_0.bam"),
        "--genome", "b37",
    ])
    allele_groups = ["num_ref", "num_alt", "total_depth"]
    for allele_group in allele_groups:
        result[allele_group] = result[allele_group].astype(int)
    eq_(cols_concat(
            result,
            ["contig", "interbase_start"] + allele_groups),
        {
            '22-50636217-0-0-0',
            '22-50875932-0-0-0',
            '22-21829554-0-0-0',
            "22-46931059-50-0-50",
            "22-46931061-51-0-51",
    })

    # Same thing but with chunk rows = 1
    with temp_file(".csv") as out_csv:
        run([
            data_path("CELSR1/vcfs/vcf_1.vcf"),
            "--include-read-evidence",
            "--reads", data_path("CELSR1/bams/bam_0.bam"),
            "--genome", "b37",
            "--chunk-rows", "1",
            "--out", out_csv,
        ])
        result = pandas.read_csv(out_csv)
    
        allele_groups = ["num_ref", "num_alt", "total_depth"]
        for allele_group in allele_groups:
            result[allele_group] = result[allele_group].astype(int)
        eq_(cols_concat(
                result,
                ["contig", "interbase_start"] + allele_groups),
            {
                '22-50636217-0-0-0',
                '22-50875932-0-0-0',
                '22-21829554-0-0-0',
                "22-46931059-50-0-50",
                "22-46931061-51-0-51",
        })

    result = run([
        "--include-read-evidence",
        "--reads", data_path("gatk_mini_bundle_extract.bam"),
        "--read-source-name", "foo",
        "--single-variant", "chr20:10008951", "C", "A",
        "--genome", "b37",
    ])
    for allele_group in allele_groups:
        result[allele_group] = result[allele_group].astype(int)
    eq_(cols_concat(result, expected_cols + allele_groups),
        {"GRCh37-20-10008950-10008951-C-A-4-1-5"})

    result = run([
        "--include-read-evidence",
        "--reads", data_path("gatk_mini_bundle_extract.bam"),
        "--read-source-name", "foo",
        "--single-variant", "chr20:10008951", "C", "A",
        "--genome", "b37",
        "--is-reverse",
    ])
    for allele_group in allele_groups:
        result[allele_group] = result[allele_group].astype(int)
    eq_(cols_concat(result, expected_cols + allele_groups),
        {"GRCh37-20-10008950-10008951-C-A-1-0-1"})


def test_filtering():
    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--genome", "b37",
        "--ref", "A",
    ])
    eq_(sorted(cols_concat(result, expected_cols)), sorted({
        "GRCh37-22-46931059-46931060-A-C",
        "GRCh37-22-50636217-50636218-A-C",
        "GRCh37-22-50875932-50875933-A-C",
    }))

    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--genome", "b37",
        "--ref", "A",
        "--variant-locus", "22:50636218",
    ])
    eq_(sorted(cols_concat(result, expected_cols)), sorted({
        "GRCh37-22-50636217-50636218-A-C",
    }))

    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        data_path("CELSR1/vcfs/vcf_2.vcf"),
        "--alt", "C", "G",
        "--genome", "b37"
    ])
    eq_(sorted(cols_concat(result, expected_cols)), sorted({
        "GRCh37-22-21829554-21829555-T-G",
        "GRCh37-22-45309892-45309893-T-G",
        "GRCh37-22-46931059-46931060-A-C",
        "GRCh37-22-50636217-50636218-A-C",
        "GRCh37-22-50875932-50875933-A-C",
    }))

'''
def test_fields():
    result = run([
        "--field",
        "foo:ref.lower()",
        "gene_names[0]",
        "--variants", data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--variant-filter", "ref=='A'",
        "--variant-genome", "b37"
    ])
    eq_(sorted(cols_concat(result, expected_cols + ["foo", "gene_names[0]"])),
        sorted({
            "GRCh37-22-46931059-46931060-A-C-a-CELSR1",
            "GRCh37-22-50636217-50636218-A-C-a-TRABD",
            "GRCh37-22-50875932-50875933-A-C-a-PPP6R2",
        }))
'''
def test_round_trip():
    with temp_file(".csv") as out_csv:
        variants.run([
            data_path("CELSR1/vcfs/vcf_1.vcf"),
            "--out", out_csv,
            "--genome", "b37",
            "--ref", "A",
            "--include-gene",
        ])
        result1 = pandas.read_csv(out_csv)
        eq_(sorted(cols_concat(
                result1, expected_cols + ["gene"])),
            sorted({
                "GRCh37-22-46931059-46931060-A-C-CELSR1",
                "GRCh37-22-50636217-50636218-A-C-TRABD",
                "GRCh37-22-50875932-50875933-A-C-PPP6R2",
            }))

        result2 = run([
            out_csv,
            "--include-gene",
        ])
        eq_(sorted(cols_concat(
                result2,
                expected_cols + ["gene"])),
            sorted({
                "GRCh37-22-46931059-46931060-A-C-CELSR1",
                "GRCh37-22-50636217-50636218-A-C-TRABD",
                "GRCh37-22-50875932-50875933-A-C-PPP6R2",
            }))

def test_distinct_variants():
    result = run([
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        data_path("CELSR1/vcfs/vcf_1.vcf"),
        "--genome", "b37",
        "--ref", "A", "T",
        "--variant-source-name", "first", "second",
    ])
    eq_(sorted(cols_concat(result, expected_cols + ["sources"])),
        sorted({
            "GRCh37-22-21829554-21829555-T-G-first second",
            "GRCh37-22-46931059-46931060-A-C-first second",
            "GRCh37-22-50636217-50636218-A-C-first second",
            "GRCh37-22-50875932-50875933-A-C-first second",
        }))




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyfaidx

def variant_context(
        reference_fasta,
        contig,
        inclusive_start,
        inclusive_end,
        alt,
        context_length):  
    """
    Retrieve the surronding reference region from a variant.

    SNVs are canonicalized so the reference base is a pyrmidine (C/T). For
    indels the reverse complement will still be taken if the first base of
    the reference is not a pyrmidine, but since the reference will also be
    reversed, that doesn't guarantee it will start with a pyrmidine.

    Parameters
    ----------
    reference_fasta : FastaReference
        reference sequence from pyfaidx package

    contig : str
        Chromosome of the variant

    inclusive_start : int
        start of the variant in 1-based inclusive coordinates

    inclusive_end : int
        end of the variant in 1-based inclusive coordinates

    alt : string
        alt sequence

    context_length : int
        number of bases on either side of the variant to return

    Returns
    ---------
    A tuple of (5', mutation, 3') where
        5' - bases immediately 5 prime to the mutation
        
        3' - bases immediately 3 prime to the mutation
        
        mutation - the ref sequence followed by a > character followed by the
            the alt sequence
    """

    # Move from 1-base coorindates to 0-base coordinates
    start = int(inclusive_start) - 1
    end = int(inclusive_end)

    full_sequence = reference_fasta[contig]

    left = str(full_sequence[start - context_length:start].seq).upper()
    middle = str(full_sequence[start: end].seq).upper()
    right = str(full_sequence[end: end + context_length].seq).upper()

    # Complement and reverse the context if necessary so the ref base is a
    # pyrmidine (C/T)
    if middle[0] in ('A', 'G'):
        context_5prime = pyfaidx.complement(right)[::-1]
        context_3prime = pyfaidx.complement(left)[::-1]
        context_mutation = "%s>%s" % (
            pyfaidx.complement(middle)[::-1], pyfaidx.complement(alt)[::-1])
    else:
        context_5prime = left
        context_3prime = right
        context_mutation = "%s>%s" % (middle, alt)

    return (context_5prime, context_mutation, context_3prime)




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import collections
from nose.tools import eq_, assert_raises

from pysam.csamfile import Samfile

from varcode import Variant as VarcodeVariant

from . import data_path

from varlens.locus import Locus
from varlens.read_evidence import PileupCollection

Variant = collections.namedtuple("Variant", "locus ref alt")

def filtered_read_names(filtered_evidence):
    assert filtered_evidence.parent is not None
    full = set(filtered_evidence.parent.read_attribute('query_name'))
    filtered = set(filtered_evidence.read_attribute('query_name'))
    assert filtered - full == set()
    return full - filtered

def test_read_evidence_rna1_single_base_loci():
    loci = [
        Locus.from_inclusive_coordinates("17", 41244936, 41244936),  # 0
        Locus.from_inclusive_coordinates("17", 41244937, 41244937),  # 1
        Locus.from_inclusive_coordinates("17", 41244935, 41244935),  # 2
        Locus.from_inclusive_coordinates("17", 41244933, 41244933),  # 3
        Locus.from_inclusive_coordinates("17", 41244853, 41244853),  # 4
        Locus.from_inclusive_coordinates("17", 41244857, 41244857),  # 5
        Locus.from_inclusive_coordinates("17", 41244864, 41244864),  # 6
        Locus.from_inclusive_coordinates("17", 41244879, 41244879),  # 7
        Locus.from_inclusive_coordinates("17", 41244901, 41244901),  # 8
        Locus.from_inclusive_coordinates("17", 41244910, 41244910),  # 9
        Locus.from_inclusive_coordinates("17", 41244917, 41244917),  # 10
        Locus.from_inclusive_coordinates("17", 41244972, 41244972),  # 11
        Locus.from_inclusive_coordinates("17", 41244973, 41244973),  # 12
        Locus.from_inclusive_coordinates("17", 41245026, 41245026),  # 13
        Locus.from_inclusive_coordinates("17", 41245027, 41245027),  # 14
        Locus.from_inclusive_coordinates("17", 41245019, 41245019),  # 15
        Locus.from_inclusive_coordinates("17", 41245018, 41245018),  # 16
    ]
    evidence = PileupCollection.from_bam(
        data_path("rna_chr17_41244936.bam"), loci)

    eq_(evidence.allele_summary(loci[0]), [("A", 11), ("G", 7)])
    eq_(evidence.allele_summary(loci[1]), [("G", 17), ("A", 1)])
    eq_(evidence.allele_summary(loci[2]), [("C", 18)])
    eq_(evidence.allele_summary(loci[3]), [("A", 17), ("G", 1)])
    eq_(evidence.allele_summary(loci[4]), [("C", 1)])
    eq_(evidence.allele_summary(loci[5]), [("T", 2)])
    eq_(evidence.allele_summary(loci[6]), [("T", 4)])
    eq_(evidence.allele_summary(loci[7]), [("C", 8)])
    eq_(evidence.allele_summary(loci[8]), [("C", 8)])
    eq_(evidence.allele_summary(loci[9]), [("C", 9)])
    eq_(evidence.allele_summary(loci[10]), [("A", 10)])
    eq_(evidence.allele_summary(loci[11]), [("T", 11)])
    eq_(evidence.allele_summary(loci[12]), [("T", 11)])
    eq_(evidence.allele_summary(loci[13]), [("C", 1)])
    eq_(evidence.allele_summary(loci[14]), [("G", 1)])
    eq_(evidence.allele_summary(loci[15]), [("T", 8)])
    eq_(evidence.allele_summary(loci[16]), [("T", 8)])

def test_read_evidence_rna1_multi_base_loci():
    loci = [
        Locus.from_inclusive_coordinates("17", 41244853, 41244854),  # 0
        Locus.from_inclusive_coordinates("17", 41244853, 41244857),  # 1
        Locus.from_inclusive_coordinates("17", 41244854, 41244857),  # 2
        Locus.from_inclusive_coordinates("17", 41244852, 41244857),  # 3
        Locus.from_inclusive_coordinates("17", 41244933, 41244936),  # 4
        Locus.from_inclusive_coordinates("17", 41244933, 41244937),  # 5
        Locus.from_inclusive_coordinates("17", 41244971, 41244973),  # 6
        Locus.from_inclusive_coordinates("17", 41265063, 41265067),  # 7
    ]
    evidence = PileupCollection.from_bam(
        data_path("rna_chr17_41244936.bam"), loci)
    eq_(evidence.allele_summary(loci[0]), [("CT", 1)])
    eq_(evidence.allele_summary(loci[1]), [("CTTTT", 1)])
    eq_(evidence.allele_summary(loci[2]), [("TTTT", 1)])
    eq_(evidence.allele_summary(loci[3]), [])
    eq_(evidence.allele_summary(loci[4]),
        [("AACA", 11), ("AACG", 6), ("GACG", 1)])
    eq_(evidence.allele_summary(loci[5]),
        [("AACAG", 10), ("AACGG", 6), ("AACAA", 1), ("GACGG", 1)])
    eq_(evidence.allele_summary(loci[6]), [("ATT", 11)])
    eq_(evidence.allele_summary(loci[7]), [("ACCCG", 1)])

def test_read_evidence_gatk_mini_bundle_extract():
    loci = [
        Locus.from_inclusive_coordinates("20", 9999996, 9999996),    # 0
        Locus.from_inclusive_coordinates("20", 10260442),            # 1
        Locus.from_inclusive_coordinates("20", 10006823),            # 2
        Locus.from_inclusive_coordinates("20", 10006819, 10006823),  # 3
        Locus.from_inclusive_coordinates("20", 10006819, 10006825),  # 4
        Locus.from_inclusive_coordinates("20", 10006822, 10006827),  # 5
        Locus.from_inclusive_coordinates("20", 10007175),            # 6
        Locus.from_inclusive_coordinates("20", 10007174, 10007176),  # 7
        Locus.from_inclusive_coordinates("20", 1, 3),                # 8
        Locus.from_inclusive_coordinates("20", 10008796),            # 9
        Locus.from_inclusive_coordinates("20", 10008921),            # 10
    ]
    handle = Samfile(data_path("gatk_mini_bundle_extract.bam"))
    evidence = PileupCollection.from_bam(handle, loci)

    eq_(evidence.allele_summary(loci[0]), [("ACT", 9)])
    eq_(evidence.filter(drop_duplicates=True).allele_summary(loci[0]),
        [("ACT", 8)])
    eq_(evidence.allele_summary(loci[1]), [("T", 7)])
    eq_(evidence.filter().allele_summary(loci[2]), [("", 6), ("C", 2)])
    eq_(evidence.filter(
        drop_duplicates=True, min_base_quality=50).allele_summary(loci[2]),
        [])
    eq_(evidence.filter(drop_duplicates=True).allele_summary(loci[2]),
        [("", 5), ("C", 1)])
    eq_(evidence.filter(
        drop_duplicates=True, min_mapping_quality=60).allele_summary(
            loci[2]),
        [("", 5), ("C", 1)])
    eq_(evidence.filter(drop_duplicates=True,
        min_mapping_quality=61).allele_summary(loci[2]), [("", 2)])
    eq_(evidence.filter(drop_duplicates=True,
        min_mapping_quality=61).allele_summary(loci[3]), [("A", 2)])
    eq_(evidence.filter(drop_duplicates=True,
        min_mapping_quality=61).allele_summary(loci[4]), [("AAA", 2)])
    eq_(evidence.filter(drop_duplicates=True,
        min_mapping_quality=61).allele_summary(loci[5]), [("AAAC", 2)])
    eq_(evidence.filter().allele_summary(loci[6]), [("T", 5), ("C", 3)])
    eq_(evidence.filter(min_base_quality=30).allele_summary(loci[6]),
        [("T", 4), ("C", 3)])
    eq_(evidence.filter().allele_summary(loci[7]),
        [("CTT", 5), ("CCT", 3)])
    eq_(evidence.filter(min_base_quality=30).allele_summary(loci[7]),
        [("CTT", 3), ("CCT", 2)])
    eq_(evidence.filter(min_base_quality=32).allele_summary(loci[2]),
        [("", 6), ("C", 1)])
    eq_(filtered_read_names(evidence.at(loci[2]).filter(min_base_quality=32)),
        {'20GAVAAXX100126:4:3:18352:43857'})
    eq_(evidence.allele_summary(loci[8]), [])
    eq_(evidence.filter(drop_duplicates=True).allele_summary(loci[8]), [])
    assert_raises(KeyError,
        evidence.allele_summary,
        Locus.from_inclusive_coordinates("20", 10009174, 10009176))
    eq_(filtered_read_names(
            evidence.at(loci[9]).filter(drop_improper_mate_pairs=True)),
        {'20FUKAAXX100202:8:68:1530:49310'})
    eq_(len(evidence.at(loci[8]).read_attribute('mapping_quality')), 0)
    eq_(list(evidence.at(loci[9]).read_attribute('mapping_quality')),
        list(evidence.at(loci[9]).read_attributes().mapping_quality))
    eq_(evidence.filter(drop_duplicates=True).allele_summary(loci[10]),
        [('C', 2), ('CA', 1), ('CAA', 1)])
    eq_(evidence.filter(drop_duplicates=True).allele_summary(
            Locus.from_interbase_coordinates(
                loci[10].contig, loci[10].start, loci[10].start)),
        [('', 2), ('A', 1), ('AA', 1)])


def test_read_evidence_variant_matching_gatk_mini_bundle_extract():
    handle = Samfile(data_path("gatk_mini_bundle_extract.bam"))

    loci = [
        Locus.from_inclusive_coordinates("20", 10008951),  # 0
        Locus.from_inclusive_coordinates("20", 10009053),  # 1
        Locus.from_inclusive_coordinates("20", 10009053, 10009054),  # 2
        Locus.from_inclusive_coordinates("20", 10006822),  # 3
        Locus.from_inclusive_coordinates("20", 10006822, 10006823),  # 4

    ]
    evidence = PileupCollection.from_bam(handle, loci)

    eq_(evidence.match_summary(Variant(loci[0], "A", "C")),
        [('A', 1), ('C', 4)])
    eq_(evidence.filter(drop_duplicates=True).match_summary(
            Variant(loci[0], "A", "C")),
        [('A', 0), ('C', 3)])
    eq_(evidence.match_summary(Variant(loci[1], "A", "C")),
        [('A', 3), ('C', 0)])
    eq_(evidence.match_summary(Variant(loci[1], "A", "CC")),
        [('A', 3), ('CC', 0)])
    eq_(evidence.match_summary(Variant(loci[1], "A", "")),
        [('A', 3), ('', 0)])
    eq_(evidence.match_summary(Variant(loci[1], "A", "")),
        [('A', 3), ('', 0)])
    eq_(evidence.match_summary(Variant(loci[2], "AT", "")),
        [('AT', 3), ('', 0)])
    eq_(evidence.match_summary(Variant(loci[3], "A", "")),
        [('A', 2), ('', 6)])
    eq_(evidence.match_summary(Variant(loci[4], "AC", "")),
        [('AC', 2), ('', 6)])
    eq_(evidence.match_summary(
            Variant(loci[4], "AC", ""),
            lambda e: e.read_attributes().mapping_quality.mean()),
        [('AC', 60.0), ('', 65.0)])

def test_read_evidence_variant_matching_gatk_bundle_native_varcode_variant():
    # Try native varcode Variant.
    handle = Samfile(data_path("gatk_mini_bundle_extract.bam"))
    locus = Locus.from_inclusive_coordinates("20", 10008951)
    variant = VarcodeVariant(
        locus.contig,
        locus.position + 1,  # inclusive not interbase
        "A",
        "C")
    evidence = PileupCollection.from_bam(handle, [variant])
    eq_(evidence.match_summary(variant),
        [('A', 1), ('C', 4)])


def test_read_evidence_variant_matching_gatk_mini_bundle_extract_warning():
    filename = data_path("gatk_mini_bundle_extract.bam")

    # Should log a warning but pass.
    loci = [
        Locus.from_inclusive_coordinates("20", 10009053, 10009054),  # 0
    ]
    evidence = PileupCollection.from_bam(filename, loci)
    eq_(evidence.match_summary(Variant(loci[0], "A", "")),
        [('A', 0), ('', 0), ('AT', 3)])





# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

def alignment_key(pysam_alignment_record):
    '''
    Return the identifying attributes of a `pysam.AlignedSegment` instance.
    This is necessary since these objects do not support a useful notion of
    equality (they compare on identify by default).
    '''
    return (
        read_key(pysam_alignment_record),
        pysam_alignment_record.query_alignment_start,
        pysam_alignment_record.query_alignment_end,
    )

def read_key(pysam_alignment_record):
    '''
    Given a `pysam.AlignedSegment` instance, return the attributes identifying
    the *read* it comes from (not the alignment). There may be more than one
    alignment for a read, e.g. chimeric and secondary alignments.
    '''
    return (
        pysam_alignment_record.query_name,
        pysam_alignment_record.is_duplicate,
        pysam_alignment_record.is_read1,
        pysam_alignment_record.is_read2,
    )



# Copyright (c) 2014. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import os

from setuptools import setup

current_directory = os.path.dirname(__file__)

if __name__ == '__main__':
    setup(
        name='varlens',
        packages=["varlens", "varlens.commands", "varlens.read_evidence"],
        version="0.0.2",
        description=(
            "commandline manipulation of genomic variants and NGS reads"),
        long_description=open('README.rst').read(),
        url="https://github.com/hammerlab/varlens",
        author="Tim O'Donnell",
        author_email="timodonnell@gmail.com",
        license="http://www.apache.org/licenses/LICENSE-2.0.html",
        entry_points={
            'console_scripts': [
                'varlens-allele-support = varlens.commands.allele_support:run',
                'varlens-variants = varlens.commands.variants:run',
                'varlens-reads = varlens.commands.reads:run',
            ],
        },
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
        install_requires=[
            'numpy',
            'intervaltree',
            'pysam>=0.8.4',
            'typechecks',
            'varcode',
            'pyfaidx',
            'mhctools',
            'topiary',
        ],
    )



'''
Given one or more BAMs and some genomic sites to consider, write a csv file
giving counts of reads supporting each allele at each site for each BAM.

The genomic sites to consider may be specified by locus (--locus option), or via
one or more VCF files.

The positions outputted by this command are in *interbase coordinates*, i.e.
starting at 0, inclusive on first index, exclusive on second (as opposed to
the one-based inclusive coordinates used in VCF files).

Example:

%(prog)s \\
    --reads test/data/CELSR1/bams/bam_1.bam \\
    --locus 22:46931061 22:46931063

'''
from __future__ import absolute_import

import argparse
import csv
import sys
import logging

from .. import loci_util
from .. import reads_util
from .. import variants_util

from . import configure_logging
from .. import support
from ..read_evidence.pileup_collection import to_locus

parser = argparse.ArgumentParser(usage=__doc__)
group = parser.add_argument_group("output arguments")
group.add_argument("--out")
group.add_argument("-v", "--verbose", action="store_true", default=False)
loci_util.add_args(parser.add_argument_group("loci specification"))
variants_util.add_args(parser)
reads_util.add_args(parser)

def run(raw_args=sys.argv[1:]):
    args = parser.parse_args(raw_args)
    configure_logging(args)

    loci = loci_util.load_from_args(args)  # may be None
    variants_df = variants_util.load_from_args_as_dataframe(args)
    if variants_df is not None:
        variant_loci = loci_util.Loci(
            to_locus(variant)
            for variant in variants_df["variant"])
        loci = variant_loci if loci is None else loci.union(variant_loci)
    
    if not loci:
        if variants_df is not None:
            parser.error("No loci: variants specified but none remained "
                "after filtering")
        else:
            parser.error("No genomic loci or variants specified.")

    logging.info("Loaded %d genomic loci." % len(loci))

    read_sources = reads_util.load_from_args(args)

    if read_sources is None:
        parser.error("No read sources (--reads argument) specified.")

    out_fd = open(args.out, "w") if args.out else sys.stdout
    writer = csv.writer(out_fd)

    rows_generator = support.allele_support_rows(loci, read_sources)
    for (i, row) in enumerate(rows_generator):
        if i == 0:
            writer.writerow(row.index.tolist())
        writer.writerow([str(x) for x in row])

    if out_fd is not sys.stdout:
        out_fd.close()
        print("Wrote: %s" % args.out)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from collections import namedtuple, defaultdict, OrderedDict
import logging
import itertools
import re

import pandas
from pysam.csamfile import Samfile
import typechecks
import pyensembl

from ..locus import Locus
from . import Pileup, PileupElement, alignment_key, read_key

MatchingEvidence = namedtuple("MatchingEvidence", "ref alt other")

class PileupCollection(object):
    '''
    A collection of Pileup instances for some loci.

    Attributes
    ----------
    pileups (optional) : dict of Locus -> Pileup
        A map from 1-base Locus instances to the Pileup at that locus.

    parent (optional) : PileupCollection
        For PileupCollection instances that are derived (e.g. by filtering)
        from another instance, this attribute tracks the parent instance.
    '''
    def __init__(self, pileups=None, parent=None):
        '''
        Construct a new PileupCollection.
        '''
        if pileups is None:
            pileups = {}
        self.pileups = pileups
        self.parent = parent

    def pileup(self, locus):
        '''
        Given a 1-base locus, return the Pileup at that locus.

        Raises a KeyError if this PileupCollection does not have a Pileup at
        the specified locus.
        '''
        locus = to_locus(locus)
        if len(locus.positions) != 1:
            raise ValueError("Not a single-base locus: %s" % locus)
        return self.pileups[locus]

    def at(self, *loci):
        '''
        Return a new PileupCollection instance including only pileups for 
        the specified loci.
        '''
        loci = [to_locus(obj) for obj in loci]
        single_position_loci = []
        for locus in loci:
            for position in locus.positions:
                single_position_loci.append(
                    Locus.from_interbase_coordinates(locus.contig, position))
        pileups = dict(
            (locus, self.pileups[locus]) for locus in single_position_loci)
        return PileupCollection(pileups, self)

    def loci(self):
        '''
        Returns the loci included in this instance.
        '''
        return list(self.pileups)

    def reads(self):
        '''
        The reads in this PileupCollection. All reads will have an alignment
        that overlaps at least one of the included loci.

        Since SAM (and pysam) have no real notion of a "read", the returned
        instances are actually pysam.AlignedSegment instances, (i.e. 
        alignments). However, only one alignment will be returned by this
        method per read.
        
        Returns
        ----------
        List of pysam.AlignedSegment instances. If a particular read has more
        than one alignment in this PileupCollection (e.g. one primary and one
        secondary), then the alignment returned is the one with the highest
        mapping quality.
        '''
        # TODO: Optimize this.

        def alignment_precedence(pysam_alignment_record):
            return pysam_alignment_record.mapping_quality

        result = {}
        for pileup in self.pileups.values():
            for e in pileup.elements:
                key = read_key(e.alignment)
                if key not in result or (
                        alignment_precedence(e.alignment) >
                        alignment_precedence(result[key])):
                    result[key] = e.alignment
        return list(result.values())

    def num_reads(self):
        '''
        Returns the number of reads in this PileupCollection.
        '''
        return len(self.reads())

    def read_attribute(self, attribute):
        '''
        Query a read attribute across all reads in this PileupCollection.

        Parameters
        ----------
        attribute : string
            Attribute to query. See `PileupCollection.read_attributes` for
            possible attributes.

        Returns
        ----------
        pandas.Series instance giving values for the attribute in each read.
        The order of reads is fixed, so multiple calls to this function will
        return corresponding values. 
        '''
        return self.read_attributes([attribute])[attribute]

    def read_attributes(self, attributes=None):
        '''
        Collect read attributes across reads in this PileupCollection into a
        pandas.DataFrame.

        Valid attributes are the following properties of a pysam.AlignedSegment
        instance. See:
            http://pysam.readthedocs.org/en/latest/api.html#pysam.AlignedSegment
        for the meaning of these attributes.

             * cigarstring
             * flag
             * inferred_length
             * is_duplicate
             * is_paired
             * is_proper_pair
             * is_qcfail
             * is_read1
             * is_read2
             * is_reverse
             * is_secondary
             * is_unmapped
             * mapping_quality
             * mate_is_reverse
             * mate_is_unmapped
             * next_reference_id
             * next_reference_start
             * query_alignment_end
             * query_alignment_length
             * query_alignment_qualities
             * query_alignment_sequence
             * query_alignment_start
             * query_length
             * query_name
             * reference_end
             * reference_id
             * reference_length
             * reference_start
             * template_length

        (Note: the above list is parsed into the _READ_ATTRIBUTE_NAMES class
        variable, so be careful when modifying it.)

        Additionally, for alignment "tags" (arbitrary key values associated
        with an alignment), a column of the form "TAG_{tag name}" is
        included.

        Finally, the column "pysam_alignment_record" gives the underlying
        `pysam.AlignedSegment` instances.

        Parameters
        ----------
        attributes (optional): list of strings
            List of columns to include. If unspecified, all columns are
            included in the result.

        Returns
        ----------
        pandas.DataFrame of read attributes.

        '''
        def include(attribute):
            return attributes is None or attribute in attributes

        reads = self.reads()

        possible_column_names = list(PileupCollection._READ_ATTRIBUTE_NAMES)
        result = OrderedDict(
            (name, [getattr(read, name) for read in reads])
            for name in PileupCollection._READ_ATTRIBUTE_NAMES
            if include(name))

        # Add tag columns.
        if reads:
            tag_dicts = [dict(x.get_tags()) for x in reads]
            tag_keys = set.union(
                *[set(item.keys()) for item in tag_dicts])
            for tag_key in sorted(tag_keys):
                column_name = "TAG_%s" % tag_key
                possible_column_names.append(column_name)
                if include(column_name):
                    result[column_name] = [d.get(tag_key) for d in tag_dicts]

        # Lastly, we include the underlying pysam alignment record.
        possible_column_names.append("pysam_alignment_record")
        if include("pysam_alignment_record"):
            result["pysam_alignment_record"] = reads

        # If particular attributes were requested, check that they're here.
        if attributes is not None:
            for attribute in attributes:
                if attribute not in result:
                    raise ValueError(
                        "No such attribute: %s. Valid attributes are: %s"
                        % (attribute, " ".join(possible_column_names)))
            assert set(attributes) == set(result)

        return pandas.DataFrame(result)

    # Parse the bulleted list of attributes in the previous method's docstring.
    _READ_ATTRIBUTE_NAMES = []
    for line in read_attributes.__doc__.split("\n"):
        match = re.match("\s+\*\s*(\w+)\s*", line)
        if match:
            _READ_ATTRIBUTE_NAMES.append(match.groups()[0])

    def group_by_allele(self, locus):
        '''
        Split the PileupCollection by the alleles suggested by the reads at the
        specified locus.

        If a read has an insertion immediately following the locus, then the
        insertion is included in the allele. For example, if locus is the
        1-base range [5,6), one allele might be "AGA", indicating that at
        locus 5 some read has an "A" followed by a 2-base insertion ("GA"). If
        a read has a deletion at the specified locus, the allele is the empty
        string.

        The given locus may include any number of bases. If the locus includes
        multiple bases, then the alleles consist of all bases aligning to that
        range in any read. Note that only sequences actually sequenced in a
        particular read are included. For example, if one read has "ATT" at a
        locus and another read has "GCC", then the alleles are "ATT" and
        "GCC", but not "GTT". That is, the bases in each allele are phased. For
        this reason, only reads that overlap the entire locus are included.

        If the locus is an empty interval (e.g. [5,5) ), then the alleles
        consist only of inserted bases. In this example, only bases inserted
        immediately after locus 5 would be included (but *not* the base
        actually at locus 5). In the previous insertion example, the allele
        would be "GA", indicating a 2-base insertion. Reads that have no
        insertion at that position (matches or deletions) would have the empty
        string as their allele.

        Parameters
        ----------
        locus : Locus
            The reference locus, encompassing 0 or more bases.

        Returns
        ----------
        A dict of string -> PileupCollection. The keys are nucleotide strings
        giving the bases sequenced at the locus, and the values are
        PileupCollection instances of the alignments that support that allele.

        '''
        locus = to_locus(locus)
        read_to_allele = None
        loci = []
        if locus.positions:
            # Our locus includes at least one reference base.
            for position in locus.positions:
                base_position = Locus.from_interbase_coordinates(
                    locus.contig, position)
                loci.append(base_position)
                new_read_to_allele = {}
                for element in self.pileups[base_position]:
                    allele_prefix = ""
                    key = alignment_key(element.alignment)
                    if read_to_allele is not None:
                        try:
                            allele_prefix = read_to_allele[key]
                        except KeyError:
                            continue
                    allele = allele_prefix + element.bases
                    new_read_to_allele[key] = allele
                read_to_allele = new_read_to_allele
        else:
            # Our locus is between reference bases.
            position_before = Locus.from_interbase_coordinates(
                locus.contig, locus.start)
            loci.append(position_before)
            read_to_allele = {}
            for element in self.pileups[position_before]:
                allele = element.bases[1:]
                read_to_allele[alignment_key(element.alignment)] = allele

        split = defaultdict(lambda: PileupCollection(pileups={}, parent=self))
        for locus in loci:
            pileup = self.pileups[locus]
            for e in pileup.elements:
                key = read_to_allele.get(alignment_key(e.alignment))
                if key is not None:
                    if locus in split[key].pileups:
                        split[key].pileups[locus].append(e)
                    else:
                        split[key].pileups[locus] = Pileup(locus, [e])

        # Sort by number of reads (descending). Break ties with the
        # lexicographic ordering of the allele string.
        def sorter(pair):
            (allele, pileup_collection) = pair
            return (-1 * pileup_collection.num_reads(), allele)
        return OrderedDict(sorted(split.items(), key=sorter))

    def allele_summary(self, locus, score=lambda x: x.num_reads()):
        '''
        Convenience method to summarize the evidence for each of the alleles
        present at a locus. Applies a score function to the PileupCollection
        associated with each allele.

        See also `PileupCollection.group_by_allele`.

        Parameters
        ----------
        locus : Locus
            The reference locus, encompassing 0 or more bases.

        score (optional) : PileupCollection -> object
            Function to apply to summarize the evidence for each allele.
            Default: count number of reads.

        Returns
        ----------
        List of (allele, score) pairs.
        '''
        locus = to_locus(locus)
        return [
            (allele, score(x))
            for (allele, x) in self.group_by_allele(locus).items()
        ]

    def group_by_match(self, variant):
        '''
        Given a variant, split the PileupCollection based on whether it the
        data supports the reference allele, the alternate allele, or neither.

        Parameters
        ----------
        variant : Variant
            The variant. Must have fields 'locus', 'ref', and 'alt'.

        Returns
        ----------
        A MatchingEvidence named tuple with fields (ref, alt, other),
        each of which is a string -> PileupCollection dict mapping alleles
        to the PileupCollection of evidence supporting them.
        '''
        locus = to_locus(variant)
        if len(variant.ref) != len(locus.positions):
            logging.warning(
                "Ref is length %d but locus has %d bases in variant: %s" %
                (len(variant.ref), len(locus.positions), str(variant)))

        alleles_dict = self.group_by_allele(locus)
        single_base_loci = [
            Locus.from_interbase_coordinates(locus.contig, position)
            for position in locus.positions
        ]
        empty_pileups = dict(
            (locus, Pileup(locus=locus, elements=[]))
            for locus in single_base_loci)
        empty_collection = PileupCollection(pileups=empty_pileups, parent=self)

        ref = {variant.ref: alleles_dict.pop(variant.ref, empty_collection)}
        alt = {variant.alt: alleles_dict.pop(variant.alt, empty_collection)}
        other = alleles_dict

        # TODO: consider end of read issues for insertions
        return MatchingEvidence(ref, alt, other)

    def match_summary(self, variant, score=lambda x: x.num_reads()):
        '''
        Convenience method to summarize the evidence for and against a variant
        using a user-specified score function.

        See also `PileupCollection.group_by_match`.

        Parameters
        ----------
        variant : Variant
            The variant. Must have fields 'locus', 'ref', and 'alt'.

        score (optional) : PileupCollection -> object
            Function to apply to summarize the evidence for each allele.
            Default: count number of reads.

        Returns
        ----------
        List of (allele, score) pairs. This list will always have at least two
        elements. The first pair in the list is the reference allele. The
        second pair is the alternate. The subsequent items give the "third"
        alleles (neither ref nor alt), if any.
        '''
        split = self.group_by_match(variant)

        def name(allele_to_pileup_collection):
            return ",".join(allele_to_pileup_collection)

        def aggregate_and_score(pileup_collections):
            merged = PileupCollection.merge(*pileup_collections)
            return score(merged)

        result = [
            (name(split.ref), aggregate_and_score(split.ref.values())),
            (name(split.alt), aggregate_and_score(split.alt.values())),
        ]
        result.extend(
            (allele, score(collection))
            for (allele, collection) in split.other.items())
        return result

    def filter(self,
            drop_duplicates=False,
            drop_improper_mate_pairs=False,
            min_mapping_quality=None,
            min_base_quality=None,
            filters=None):
        '''
        Return a new PileupCollection that includes only pileup elements
        satisfying the specified criteria.

        Parameters
        ----------
        drop_duplicates (optional, default False) : boolean
            Remove alignments with the is_duplicate flag.

        drop_improper_mate_pairs (optional, default False) : boolean
            Retain only alignments that have mapped mate pairs, where one
            alignment in the pair is on the forward strand and the other
            is on the reverse.

        min_mapping_quality (optional) : int
            If specified, retain only alignments with mapping quality >= the
            specified threshold.

        min_base_quality (optional) : int
            If specified, retain only pileup elements where the base quality
            for the bases aligning to the pileup's locus are >= the specified
            threshold.

        filters (optional) : list of PileupElement -> bool functions
            User-specified filter functions to apply. This will be called on 
            each PileupElement, and should return True if the element should
            be retained.

        Returns
        ----------
        A new PileupCollection that includes the subset of the pileup elements
        matching all the specified filters.
        '''

        if filters is None:
            filters = []
        if drop_duplicates:
            filters.append(lambda e: not e.alignment.is_duplicate)
        if drop_improper_mate_pairs:
            filters.append(lambda e: e.alignment.is_proper_pair)
        if min_mapping_quality is not None:
            filters.append(
                lambda e: e.alignment.mapping_quality >= min_mapping_quality)
        if min_base_quality is not None:
            filters.append(
                lambda e: e.min_base_quality >= min_base_quality)
        pileups = OrderedDict(
            (locus, pileup.filter(filters))
            for (locus, pileup)
            in self.pileups.items())
        return PileupCollection(pileups=pileups, parent=self)

    def merge(self, *others):
        '''
        Return a new PileupCollection that is the union of self and the other
        specified collections.
        '''
        new_pileups = {}
        for collection in (self,) + others:
            for (locus, pileup) in collection.pileups.items():
                if locus in new_pileups:
                    new_pileups[locus].update(pileup)
                else:
                    new_pileups[locus] = Pileup(locus, pileup.elements)
        return PileupCollection(new_pileups, parent=self)

    @staticmethod
    def from_bam(pysam_samfile, loci, normalized_contig_names=True):
        '''
        Create a PileupCollection for a set of loci from a BAM file.

        Parameters
        ----------
        pysam_samfile : `pysam.csamfile.Samfile` instance, or filename string
            to a BAM file. The BAM file must be indexed.

        loci : list of Locus instances
            Loci to collect pileups for.

        normalized_contig_names : whether the contig names have been normalized
            (e.g. pyensembl removes the 'chr' prefix). Set to true to
            de-normalize the names when querying the BAM file.

        Returns
        ----------
        PileupCollection instance containing pileups for the specified loci.
        All alignments in the BAM file are included (e.g. duplicate reads,
        secondary alignments, etc.). See `PileupCollection.filter` if these
        need to be removed. 
        '''

        loci = [to_locus(obj) for obj in loci]

        close_on_completion = False
        if typechecks.is_string(pysam_samfile):
            pysam_samfile = Samfile(pysam_samfile)
            close_on_completion = True        

        try:
            # Map from pyensembl normalized chromosome names used in Variant to
            # the names used in the BAM file.
            if normalized_contig_names:
                chromosome_name_map = {}
                for name in pysam_samfile.references:
                    normalized = pyensembl.locus.normalize_chromosome(name)
                    chromosome_name_map[normalized] = name
                    chromosome_name_map[name] = name
            else:
                chromosome_name_map = None

            result = PileupCollection({})

            # Optimization: we sort variants so our BAM reads are localized.
            locus_iterator = itertools.chain.from_iterable(
                (Locus.from_interbase_coordinates(locus_interval.contig, pos)
                    for pos
                    in locus_interval.positions)
                for locus_interval in sorted(loci))
            for locus in locus_iterator:
                result.pileups[locus] = Pileup(locus, [])
                if normalized_contig_names:
                    try:
                        chromosome = chromosome_name_map[locus.contig]
                    except KeyError:
                        logging.warn("No such contig in bam: %s" % locus.contig)
                        continue
                else:
                    chromosome = locus.contig
                columns = pysam_samfile.pileup(
                    chromosome,
                    locus.position,
                    locus.position + 1,  # exclusive, 0-indexed
                    truncate=True,
                    stepper="nofilter")
                try:
                    column = next(columns)
                except StopIteration:
                    # No reads align to this locus.
                    continue

                # Note that storing the pileups here is necessary, since the
                # subsequent assertion will invalidate our column.
                pileups = column.pileups
                assert list(columns) == []  # column is invalid after this.
                for pileup_read in pileups:
                    if not pileup_read.is_refskip:
                        element = PileupElement.from_pysam_alignment(
                            locus, pileup_read)
                        result.pileups[locus].append(element)
            return result
        finally:
            if close_on_completion:
                pysam_samfile.close()

def to_locus(variant_or_locus):
    """
    Return a Locus object for a Variant instance.

    This is necessary since the read evidence module expects Variant instances
    to have a locus attribute st to a varcode.Locus instance of interbase
    genomic coordinates.  The rest of varcode uses a different Variant class,
    but will eventually be transitioned to interbase coordinates.

    See test/test_read_evidence.py for a definition of the Variant class that
    the read_evidence module is meant to work with.

    This function can be passed a regular varcode.Variant instance (with fields
    start, end, contig, etc.), a different kind of variant object that has a
    'locus' field, or a varcode.Locus. It will return a varcode.Locus instance.

    This should all get cleaned up once varcode switches to interbase
    coordinates and we standardize on a Variant class.
    """
    if isinstance(variant_or_locus, Locus):
        return variant_or_locus
    try:
        return variant_or_locus.locus
    except AttributeError:
        # IMPORTANT: if varcode someday changes from inclusive to interbase
        # coordinates, this will need to be updated.
        return Locus.from_inclusive_coordinates(
            variant_or_locus.contig,
            variant_or_locus.start,
            variant_or_locus.end)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import logging

import pandas

EXPECTED_COLUMNS = [
    "source",
    "contig",
    "interbase_start",
    "interbase_end",
    "allele",
    "count",
]

def allele_support_df(loci, sources):
    """
    Returns a DataFrame of allele counts for all given loci in the read sources
    """
    return pandas.DataFrame(
        allele_support_rows(loci, sources),
        columns=EXPECTED_COLUMNS)

def allele_support_rows(loci, sources):
    for source in sources:
        logging.info("Reading from: %s (%s)" % (source.name, source.filename))
        for locus in loci:
            grouped = dict(source.pileups([locus]).group_by_allele(locus))
            if grouped:
                items = grouped.items()
            else:
                items = [("N" * (locus.end - locus.start), None)]
            for (allele, group) in items:
                d = collections.OrderedDict([
                    ("source", source.name),
                    ("contig", locus.contig),
                    ("interbase_start", str(locus.start)),
                    ("interbase_end", str(locus.end)),
                    ("allele", allele),
                    ("count", group.num_reads() if group is not None else 0),
                ])
                yield pandas.Series(d)

def variant_support(variants, allele_support_df, ignore_missing=False):
    '''
    Collect the read evidence support for the given variants.

    Parameters
    ----------

    variants : iterable of varcode.Variant

    allele_support_df : dataframe
        Allele support dataframe, as output by the varlens-allele-support tool.
        It should have columns: source, contig, interbase_start, interbase_end,
        allele. The remaining columns are interpreted as read counts of various
        subsets of reads (e.g. all reads, non-duplicate reads, etc.)

    ignore_missing : boolean
        If True, then varaints with no allele counts will be interpreted as
        having 0 depth. If False, then an exception will be raised if any
        variants have no allele counts.

    Returns
    ----------

    A pandas.Panel4D frame with these axes:

    labels (axis=0) : the type of read being counted, i.e. the read count
        fields in allele_support_df.

    items (axis=1)  : the type of measurement (num_alt, num_ref, num_other,
        total_depth, alt_fraction, any_alt_fraction)

    major axis (axis=2) : the variants

    minor axis (axis=3) : the sources
    '''
    missing = [
        c for c in EXPECTED_COLUMNS if c not in allele_support_df.columns
    ]
    if missing:
        raise ValueError("Missing columns: %s" % " ".join(missing))

    # Ensure our start and end fields are ints.
    allele_support_df[["interbase_start", "interbase_end"]] = (
        allele_support_df[["interbase_start", "interbase_end"]].astype(int))

    sources = sorted(allele_support_df["source"].unique())

    allele_support_dict = collections.defaultdict(dict)
    for (i, row) in allele_support_df.iterrows():
        key = (
            row['source'],
            row.contig,
            row.interbase_start,
            row.interbase_end)
        allele_support_dict[key][row.allele] = row["count"]

    # We want an exception on bad lookups, so convert to a regular dict.
    allele_support_dict = dict(allele_support_dict)

    dataframe_dicts = collections.defaultdict(
        lambda: collections.defaultdict(list))

    for variant in variants:
        for source in sources:
            key = (source, variant.contig, variant.start - 1, variant.end)
            try:
                alleles = allele_support_dict[key]
            except KeyError:
                message = (
                    "No allele counts in source %s for variant %s" % (
                        source, str(variant)))
                if ignore_missing:
                    logging.warning(message)
                    alleles = {}
                else:
                    raise ValueError(message)

            alt = alleles.get(variant.alt, 0)
            ref = alleles.get(variant.ref, 0)
            total = sum(alleles.values())

            other = total - alt - ref

            dataframe_dicts["num_alt"][source].append(alt)
            dataframe_dicts["num_ref"][source].append(ref)
            dataframe_dicts["num_other"][source].append(other)
            dataframe_dicts["total_depth"][source].append(total)
            dataframe_dicts["alt_fraction"][source].append(
                float(alt) / max(1, total))
            dataframe_dicts["any_alt_fraction"][source].append(
                float(alt + other) / max(1, total))

    dataframes = dict(
        (label, pandas.DataFrame(value, index=variants))
        for (label, value) in dataframe_dicts.items())

    return pandas.Panel(dataframes)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from . import alignment_key

class PileupElement(object):
    '''
    A PileupElement represents the segment of an alignment that aligns to a
    particular base in the reference.

    Attributes
    ----------
    locus : Varcode.Locus
        The reference locus. Must be length 1, i.e. a single base.

    offset_start : int
        0-based start offset into the alignment sequence, inclusive

    offset_end : int
        0-based end offset into the alignment sequence, exclusive

    alignment : pysam.AlignedSegment
        pysam alignment instance

    alignment_key : tuple
        value computed from the alignment instance that uniquely specifies its
        properties. Used for comparisons since pysam.AlignedSegment instances
        do not support a useful notion of equality (they compare using object
        identity). See `read_evidence.alignment_key` for the implementation of
        this key.
    '''
    def __init__(self, locus, offset_start, offset_end, alignment):
        '''
        Construct a PileupElement object.
        '''
        assert offset_end >= offset_start, \
            "offset_start=%d > offset_end=%d" % (offset_start, offset_end)
        self.locus = locus
        self.offset_start = offset_start
        self.offset_end = offset_end
        self.alignment = alignment
        self.alignment_key = alignment_key(self.alignment)

    def fields(self):
        '''
        Fields that should be considered for our notion of object equality.
        '''
        return (
            self.locus, self.offset_start, self.offset_end, self.alignment_key)

    def __eq__(self, other):
        return hasattr(other, "fields") and self.fields() == other.fields()

    def __hash__(self):
        return hash(self.fields())

    @property
    def bases(self):
        '''
        The sequenced bases in the alignment that align to this locus in the
        genome, as a string.

        Empty string in the case of a deletion. String of length > 1 if there
        is an insertion here.
        '''
        sequence = self.alignment.query_sequence
        assert self.offset_end <= len(sequence), \
            "End offset=%d > sequence length=%d. CIGAR=%s. SEQUENCE=%s" % (
                self.offset_end,
                len(sequence),
                self.alignment.cigarstring,
                sequence)
        return sequence[self.offset_start:self.offset_end]

    @property
    def base_qualities(self):
        '''
        The phred-scaled base quality scores corresponding to `self.bases`, as
        a list.
        '''
        return self.alignment.query_qualities[
            self.offset_start:self.offset_end]

    @property
    def min_base_quality(self):
        '''
        The minimum of the base qualities. In the case of a deletion, in which
        case there are no bases in this PileupElement, the minimum is taken
        over the sequenced bases immediately before and after the deletion.
        '''
        try:
            return min(self.base_qualities)
        except ValueError:
            # We are mid-deletion. We return the minimum of the adjacent bases.
            assert self.offset_start == self.offset_end
            adjacent_qualities = [
                self.alignment.query_qualities[offset]
                for offset in [self.offset_start - 1, self.offset_start]
                if 0 <= offset < len(self.alignment.query_qualities)
            ]
            return min(adjacent_qualities)

    @staticmethod
    def from_pysam_alignment(locus, pileup_read):
        '''
        Factory function to create a new PileupElement from a pysam
        `PileupRead`.

        Parameters
        ----------
        locus : varcode.Locus
            Reference locus for which to construct a PileupElement. Must
            include exactly one base.

        pileup_read : pysam.calignmentfile.PileupRead
            pysam PileupRead instance. Its alignment must overlap the locus.

        Returns
        ----------
        PileupElement

        '''
        assert not pileup_read.is_refskip, (
            "Can't create a PileupElement in a refskip (typically an intronic "
            "gap in an RNA alignment)")

        # Pysam has an `aligned_pairs` method that gives a list of
        # (offset, locus) pairs indicating the correspondence between bases in
        # the alignment and reference loci. Here we use that to compute
        # offset_start and offset_end.
        #
        # This is slightly tricky in the case of insertions and deletions.
        # Here are examples of the desired logic.
        #
        # Target locus = 1000
        #
        # (1) Simple case: matching bases.
        #
        # OFFSET           LOCUS
        # 0                999
        # 1                1000
        # 2                1001
        #
        # DESIRED RESULT: offset_start=1, offset_end=2.
        #
        #
        # (2) A 1 base insertion at offset 2.
        #
        # OFFSET           LOCUS
        # 0                999
        # 1                1000
        # 2                None
        # 3                1001
        #
        # DESIRED RESULT: offset_start = 1, offset_end=3.
        #
        #
        # (3) A 2 base deletion at loci 1000 and 1001.
        #
        # OFFSET           LOCUS
        # 0                999
        # None             1000
        # None             1001
        # 1                1002
        #
        # DESIRED RESULT: offset_start = 1, offset_end=1.
        #
        offset_start = None
        offset_end = len(pileup_read.alignment.query_sequence)
        # TODO: doing this with get_blocks() may be faster.
        for (offset, position) in pileup_read.alignment.aligned_pairs:
            if offset is not None and position is not None:
                if position == locus.position:
                    offset_start = offset
                elif position > locus.position:
                    offset_end = offset
                    break
        if offset_start is None:
            offset_start = offset_end
        
        assert pileup_read.is_del == (offset_end - offset_start == 0), \
            "Deletion=%s but | [%d,%d) |=%d for locus %d in: \n%s" % (
                pileup_read.is_del,
                offset_start,
                offset_end,
                offset_end - offset_start,
                locus.position,
                pileup_read.alignment.aligned_pairs)

        assert offset_end >= offset_start
        result = PileupElement(
            locus, offset_start, offset_end, pileup_read.alignment)
        return result   




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import functools

from .read_source import ReadSource
from . import util

BOOLEAN_PROPERTIES = """
is_paired is_proper_pair is_qcfail is_read1 is_read2 is_reverse is_secondary
is_unmapped mate_is_reverse mate_is_unmapped is_duplicate
""".split()

STRING_PROPERTIES = """
cigarstring query_alignment_sequence query_name
""".split()

INT_PROPERTIES = """
inferred_length mapping_quality query_alignment_length query_alignment_start
query_length reference_length reference_start template_length
""".split()

# name -> (type, help, filter function)
READ_FILTERS = collections.OrderedDict()

for prop in BOOLEAN_PROPERTIES:
    READ_FILTERS[prop] = (
        bool,
        "Only reads where %s is True" % prop,
        functools.partial(
            (lambda field_name, parsed_value, read:
                bool(getattr(read, field_name))),
            prop)
    )

    READ_FILTERS["not_" + prop] = (
        bool,
        "Only reads where %s is False" % prop,
        functools.partial(
            (lambda field_name, parsed_value, read:
                not getattr(read, field_name)),
            prop)
    )

def field_contains(field_name, parsed_value, read):
    field_value = getattr(read, field_name)
    return field_value is not None and parsed_value in field_value

for prop in STRING_PROPERTIES:
    READ_FILTERS["%s" % prop] = (
        str,
        "Only reads with the specified %s" % prop,
        functools.partial(
            (lambda field_name, parsed_value, read:
                getattr(read, field_name) == parsed_value),
            prop)
    )

    READ_FILTERS["%s_contains" % prop] = (
        str,
        "Only reads where %s contains the given string" % prop,
        functools.partial(field_contains, prop))

for prop in INT_PROPERTIES:
    READ_FILTERS["%s" % prop] = (
        int,
        "Only reads with the specified %s" % prop,
        functools.partial(
            (lambda field_name, parsed_value, read:
                getattr(read, field_name) == parsed_value),
            prop)
    )

    READ_FILTERS["min_%s" % prop] = (
        int,
        "Only reads where %s >=N" % prop,
        functools.partial(
            (lambda field_name, parsed_value, read:
                getattr(read, field_name) >= parsed_value),
            prop)
    )

    READ_FILTERS["max_%s" % prop] = (
        int,
        "Only reads where %s <=N" % prop,
        functools.partial(
            (lambda field_name, parsed_value, read:
                getattr(read, field_name) <= parsed_value),
            prop)
    )

def add_args(parser, positional=False):
    """
    Extends a commandline argument parser with arguments for specifying
    read sources.
    """
    group = parser.add_argument_group("read loading")
    group.add_argument("reads" if positional else "--reads",
        nargs="+", default=[],
        help="Paths to bam files. Any number of paths may be specified.")

    group.add_argument(
        "--read-source-name",
        nargs="+",
        help="Names for each read source. The number of names specified "
        "must match the number of bam files. If not specified, filenames are "
        "used for names.")

    # Add filters
    group = parser.add_argument_group(
        "read filtering",
        "A number of read filters are available. See the pysam "
        "documentation (http://pysam.readthedocs.org/en/latest/api.html) "
        "for details on what these fields mean. When multiple filter "
        "options are specified, reads must match *all* filters.")

    for (name, (kind, message, function)) in READ_FILTERS.items():
        extra = {}
        if kind is bool:
            extra["action"] = "store_true"
            extra["default"] = None
        elif kind is int:
            extra["type"] = int
            extra["metavar"] = "N"
        elif kind is str:
            extra["metavar"] = "STRING"
        group.add_argument("--" + name.replace("_", "-"),
            help=message,
            **extra)

def load_from_args(args):
    """
    Given parsed commandline arguments, returns a list of ReadSource objects
    """
    if not args.reads:
        return None

    if args.read_source_name:
        read_source_names = util.expand(
            args.read_source_name,
            'read_source_name',
            'read source',
            len(args.reads))
    else:
        read_source_names = util.drop_prefix(args.reads)

    filters = []
    for (name, info) in READ_FILTERS.items():
        value = getattr(args, name)
        if value is not None:
            filters.append(functools.partial(info[-1], value))

    return [
        load_bam(filename, name, filters)
        for (filename, name)
        in zip(args.reads, read_source_names)
    ]

def load_bam(filename, name=None, filters=[]):
    if not name:
        name = filename
    return ReadSource(name, filename, filters)

def flatten_header(header):
    for (group, rows) in header.items():
        for (index, row) in enumerate(rows):
            if not isinstance(row, dict):
                key_values = [(row, "")]
            else:
                key_values = row.items()
            for (key, value) in key_values:
                yield (str(group), index, str(key), str(value))
